"""Bloch wave solver"""
import importlib as imp
import numpy as np,pandas as pd,pickle5,os,glob,tifffile,mrcfile
from typing import TYPE_CHECKING, Dict, Iterable, Optional, Sequence, Union
from subprocess import check_output,Popen,PIPE
from crystals import Crystal
from utils import displayStandards as dsp           #;imp.reload(dsp)
from utils import physicsConstants as cst           #;imp.reload(cst)
from utils import glob_colors as colors,handler3D as h3d
from multislice import postprocess as pp            #;imp.reload(pp)
from scattering import structure_factor as sf       #;imp.reload(sf)
from scattering import scattering_factors as scatf  #;imp.reload(scatf)
from EDutils import viewers                         #;imp.reload(viewers)
from EDutils import utilities as ut                 #;imp.reload(ut)
from EDutils import pets as pt                      ;imp.reload(pt)
from EDutils import display as EDdisp               ;imp.reload(EDdisp)
from . import util as bloch_util                    ;imp.reload(bloch_util)
felix='%s/bin/felix' %os.path.dirname(__file__)

class Bloch:
    """
    Bloch wave simulation class

    Parameters
    ---------
    file
        structure file (3D:cif_file or wallpaper info in 2D)
    name
        optional name (used to save object see set_name)
    path
        A path for the simulation folder
    beam
        dictionary passed to :meth:`~Bloch.set_beam`
    keV
        electron wavelength (see :meth:`~Bloch.set_beam`)
    u
        beam direction [ux,uy,uz] in reciprocal space (see :meth:`~Bloch.set_beam`)
    Nmax
        max order of reflections/resolution (see :meth:`~Bloch.update_Nmax`)
    Smax
        maximum excitation error (see :meth:`~Bloch.solve`)
    solve
        solve the system by diagonalizing the Bloch matrix
    felix
        Use felix solver
    nbeams
        Number of beams to use (using felix only)
    kwargs
        arguments to be passed to :meth:`~Bloch.solve`
    """
    def __init__(self,
        cif_file:str,name:str='',path:str='',
        beam:Optional[dict]={},keV:float=200,u:Sequence[float]=[0,0,1],
        Nmax:int=1,dmin:int=None,
        Smax:float=0.2,
        solve:bool=True,init=True,
        felix:bool=False,nbeams:int=200,
        eps:float=1,
        **kwargs,
    ):
        self.solved = False
        self.Nmax   = 0
        self.thick  = 100
        self.thicks = self._set_thicks((0,1000,1000))
        self.eps=eps

        self._set_structure(cif_file)
        beam_args={'keV':keV,'u':u}
        beam_args.update(beam)
        self.set_beam(**beam_args)
        self.set_name(name,path)
        self.nbeams=nbeams

        if solve :
            self.solve(Smax=Smax,Nmax=Nmax,dmin=dmin,**kwargs)
        else:
            if not felix and init:
                print(colors.blue+'...Nmax... '+colors.black)
                self.update_Nmax(Nmax,dmin)
                print(colors.blue+'...Excitation errors... '+colors.black)
                self._set_excitation_errors(Smax)
                print(colors.blue+'...Vg... '+colors.black)
                self._set_Vg()
        self.save()

        # if show_thicks or 't' in opts:
        #     self.show_beams_vs_thickness(strong=['I'])
        # if 's' in opts:
        #     self.save()


    def set_name(self,name='',path=''):
        """
        Set the basename for saving a Bloch obj as path+name+'.pkl'

        Parameters
        ----------
        name
            The name of the simulation(can include the full path)
        path
            The path to the simulation folder


        .. NOTE::
            By default the simulation name is '<cif><zone axis>_<keV>keV_bloch'
        """
        basefile = self.cif_file.replace('.cif','')
        if not name:
            u_str = ''.join(['%d' %np.round(u) for u in self.Kabc0])
            name='%s%s_%dkeV_bloch' %(os.path.basename(basefile),u_str,np.round(self.keV))
        if not path:path=os.path.dirname(basefile)
        self.path = path                #; print(self.path)
        self.name = name                #;print('name:',self.name)
        if self.pdb_file:
            check_output('mv %s %s' %(self.cif_file,self.path),shell=True)

    def update_Nmax(self,Nmax:int,dmin:int=None):
        """
        Update resolution/max order (lattice and Fhkl)

        Parameters
        ----------
        Nmax
            maximum h,k,l order
        dmin
            minimum resolution
        """

        if dmin and self.pdb_file:
            self.hklF,self.Fhkl = ut.gemmi_sf(self.pdb_file,dmin)
            self.Nmax=np.array(self.Fhkl.shape[0])//4#.min()
            print('Nmax:',self.Nmax)
            # gemmi='/home/tarik/Documents/git/github/gemmi/gemmi'
            # gemmi_cmd="%s sfcalc -v --dmin=%.3f --wavelength=%3.f --for=mott-bethe %s | " %(gemmi,dmin,self.lam,crys)
            # print(gemmi_cmd)
            # sed_cmd = r"sed 's/^ //' | sed 's/ /,/g' >>sf.txt"
            # sf = check_output("%s | %s " %(gemmi_cmd,sed_cmd) ,shell=True).decode()
            # df = pd.DataFrame('sf.txt',delimiter='\t',index_col=0,names=['index','A','phi'])
            # sf = check_output("rm sf.txt",shell=True).decode()
            # df['F'] = df.A*np.exp(1J*np.deg2rad(df.phi))
            # df[['h','k','l']]=list(map(lambda s:list(eval(s)),df.index))
            # df[['qx','qy','qz']] = df[['h','k','l']].values.dot(self.lat_vec0)
            # df['q'] = np.linalg.norm(df[['qx','qy','qz']],axis=1)
            # df['res']=1/df.q

        else:
            if type(Nmax) in [int,np.int64] :
                if not Nmax==self.Nmax:
                    self.Nmax=Nmax
                    idx = [i for i,x in enumerate(self.pattern[:,:3]) if all(x<0.99)]
                    self.pattern=self.pattern[idx,:]
                    self.hklF,self.Fhkl = sf.structure_factor3D(self.pattern,
                        2*np.pi*self.lat_vec,hklMax=2*self.Nmax)
        (h,k,l),(qx,qy,qz) = ut.get_lattice(self.lat_vec,self.Nmax)
        self.lattice = [(h,k,l),(qx,qy,qz)]

    def set_beam(self,
        keV:float=200,
        # lam:float=None,
        u:Sequence[float]=[0,0,1],
        K:Optional[Sequence[float]]=None,
        u0:Optional[Sequence[int]]=None,
    ):
        """ Set the beam direction and wavelength

        Parameters
        ----------
        keV
            beam wavelength in keV
        u
            beam direction
        K
            update beam direction and wavelength (overwrites keV)
        u0
            beam direction zone axis notation


        .. note::

            The order of priority for setting up the beam is K,u0,u
        """
        if isinstance(K,list) or isinstance(K,np.ndarray):
            self.k0 = np.linalg.norm(K)
            self.u  =  K/self.k0
        else:
            if isinstance(u0,list) or isinstance(u0,np.ndarray):
                u = np.array(u0).dot(self.lat_vec)
            self.u  = np.array(u)/np.linalg.norm(u)
            self.k0 = 1/cst.keV2lam(keV)

        self.lam = 1/self.k0
        self.keV = cst.lam2keV(self.lam)
        self.sig = cst.keV2sigma(self.keV)
        # self.u   = self.K/self.k0
        self.K   = self.k0*self.u
        self.Kuvw = self.lat_vec.dot(self.K)  #projection in reciprocal crystal basis
        self.Kabc = self.lat_vec0.dot(self.K) #projection in crystal basis
        self.Kabc0 = self.Kabc/(abs(self.Kabc)[abs(self.Kabc)>0.01].min()) #;print(Kabc)
        self.Kuvw0 = self.Kuvw/(abs(self.Kuvw)[abs(self.Kuvw)>0.01].min()) #;print(Kabc)

    def set_thickness(self,thick:float=100,v=True):
        """set thickness and update beams

        Parameters
        ----------
        thick : thickness in A
        """
        # print(type(thick))
        if type(thick) in [int,float] :self.thick=thick
        self._set_kinematic()
        if self.solved:
            self._set_intensities()
            if v: print('updated intensities')


    def solve(self,
        Smax:Optional[float]=None,hkl:Optional[Iterable[int]]=None,
        Nmax:Optional[int]=None,dmin:Optional[int]=None,
        beam:Optional[dict]={},dyngo_args:Optional[dict]={},
        thick:float=None,thicks:Sequence[float]=None,
        opts:str='sv0',
        felix:bool=False,
        nbeams:int=None,
    ):
        """ Diagonalize the Blochwave matrix

        Parameters
        ----------
        Smax
            maximum excitation error to be included
        hkl
            Beams to include
        Nmax
            max order of reflections/resolution (see :meth:`~Bloch.update_Nmax` )
        beam
            dictionary passed to :meth:`~Bloch.set_beam`
        dyngo_args
            dict contain dyngo type informations. If not empty, it solves  with Dyngo formulation which is slightly different.
        thick
            thickness of crystal (can be modified without resolving)
        thicks
            range of thickness [z_min,z_max,z_step]
        opts
            s(save) t(show intensities) z(show beam vs thickness) 0(Vopt0) v(verbose) H(show H)
        felix
            use felix solver
        nbeams
            Number of beams when using felix solver

        .. note ::

            If hkl is specified, Smax is not taken into account
        """
        if felix:
            self._prepare_Felix(nbeams=nbeams)
            self._run_felix() #wait='w' in opts)
            self._postprocess_felix(show_log='v' in opts)
            self._set_excitation_errors(hkl=self.df_G[['h','k','l']].values,felix=True)
            self._set_Vg(felix=True)
        else:
            if Nmax or dmin :self.update_Nmax(Nmax,dmin)
            if beam : self.set_beam(**beam)
            if Smax or isinstance(hkl,np.ndarray):
                self._set_excitation_errors(Smax,hkl)
                self._set_Vg()
            self._solve_Bloch(show_H='H' in opts,Vopt0='0' in opts,v='v' in opts,
                dyngo_args=dyngo_args)
        ##### postprocess
        if thick or 't' in opts:
            self.set_thickness(thick)
        if not type(thicks)==type(None) and 'z' in opts:
            self._set_beams_vs_thickness(thicks)#;print('thicks solved')
        if 's' in opts:
            self.save()

    def update(self,keV=None,u=None,Smax=None,Nmax=None,dmin=None,
            gemmi=False,hkl=None,**kwargs):
        if not gemmi:dmin=None
        self.set_beam(keV=keV,u=u)
        if Nmax or dmin :self.update_Nmax(Nmax,dmin)
        if Smax or isinstance(hkl,np.ndarray):
            self._set_excitation_errors(Smax,hkl)
            self._set_Vg()
        self.save()

    ################################################################################
    #### private
    ################################################################################
    def _prepare_Felix(self,npx=20,nbeams=None,thicks=(0,0,20),show_log=False):
        """"""
        if not nbeams:nbeams=self.nbeams
        print(colors.blue+"preparing simulation"+colors.black)
        cmd="""cd %s;
        if [ ! -d felix ];then mkdir felix;fi;
        """ %(self.path) #,os.path.realpath(felix_cif))
        # cp %s felix/felix.cif;
        # mv felix.inp felix;
        # p=Popen(cmd,shell=True)#;p.wait();o,e = p.communicate();if e:print(e)
        p=check_output(cmd,shell=True).decode();print(p)
        bloch_util.get_inp(npx,nbeams,self.u,self.keV,thicks,out=os.path.join(self.path,'felix','felix.inp'))
        ut.crys2felix(self.crys,opt='w',out=os.path.join(self.path,'felix','felix.cif'))

    def _run_felix(self,wait=True):
        if not os.path.exists(felix):
            print('Error : Running with Felix not available. You need to install felix in %s :' )
            print(felix)
            return

        print(colors.blue+"... running felix ..."+colors.black)
        cmd="""cd %s;
        cd felix;
        %s > felix.log 2>&1;
        """ %(self.path,felix)#os.path.dirname(__file__))
        p=Popen(cmd,shell=True,stdout=PIPE,stderr=PIPE);
        if wait:
            p.wait();print(p.communicate())
        else:
            return p


    def _postprocess_felix(self,show_log=False):
        if show_log:
            print(colors.blue+"felix output"+colors.black)
            with open('%s/felix/felix.log' %self.path,'r') as f:print('\n'.join(f.readlines()))

        eigvals = os.path.join(self.path,'felix/eigenvals.txt')
        check_output("sed -i -E 's/ {1,}/,/g' %s;sed -i -E 's/^,//' %s" %(eigvals,eigvals) ,shell=True)
        df = pd.read_csv(eigvals,names=['h','k','l','gr','gi'])
        self.gammaj = (df.gr+1J*df.gi).values
        # g = np.loadtxt(os.path.join(self.path,'felix/eigenvals.txt'))
        # self.gammaj = g[:,3::2]+1J*g[:,4::2];g=g[:,0]
        C = np.loadtxt(os.path.join(self.path,'felix/eigenvec.txt'))
        self.CjG = C[:,3::2]+1J*C[:,4::2]
        self.invCjG=np.conj(self.CjG.T)

        self.df_G = df[['h','k','l']]
        self.df_G.index = [str(tuple(h)) for h in self.df_G.values]
        self.solved=True

    def _solve_Bloch(self,show_H=False,Vopt0=True,dyngo_args={},v=False):
        ''' Diagonalize the Hamiltonian'''
        # Ug is a (2*Nmax+1)^3 tensor :
        # Ug[l] = [U(-N,-N,l) .. U(-N,0,l) .. U(-N,N,l)
        #          U( 0,-N,l) .. U( 0,0,l) .. U( 0,N,l)
        #          U(-N,-N,l) .. U( N,0,l) .. U( N,N,l)]
        self.dyngo = any(dyngo_args)

        hkl = self.df_G[['h','k','l']].values
        Sg  = self.df_G.Sw.values
        if self.dyngo:
            Rmat=dyngo_args['Rmat']
            self.scale=dyngo_args['scale']
            surf_norm = Rmat.dot([0,0,1])
            gn = Rmat.dot(self.df_G[['qx','qy','qz']].T).T.dot(surf_norm)
            Knorm = self.k0*surf_norm[-1]
            self.Knorm=Knorm
            sqrtkg=np.sqrt(1+gn/Knorm)  #dyngo implementation
            Sg*=2*self.k0/sqrtkg

        pre = 1/np.sqrt(1-cst.keV2v(self.keV)**2)
        Ug = pre*self.Fhkl/(self.crys.volume*np.pi)*self.eps #/3

        #####################
        # setting average potential to 0
        # Ug[U0_idx] = U0
        # and Ug(iG,jG) are obtained from Ug[h,k,l] where h,k,l = hlk_iG-hkl_jG
        #####################
        U0_idx = [2*self.Nmax]*3
        Ug[tuple(U0_idx)] = 0
        # if Vopt0 :Ug[tuple(U0_idx)] = 0   #setting average potential to 0

        print(Ug.shape)
        if v:print(colors.blue+'...assembling %dx%d matrix...' %((Sg.shape[0],)*2)+colors.black)
        H = np.diag(Sg+0J)
        for iG,hkl_G in enumerate(hkl) :
            U_iG = np.array([Ug[tuple(hkl_J+U0_idx)] for hkl_J in hkl_G-hkl]) #;print(V_iG.shape)
            # print('G : ',hkl_G)
            # print('U_G:',Ug[tuple(U0_idx+hkl_G)])
            # print('idx iG:',[tuple(hkl_J+U0_idx) for hkl_J in hkl_G-hkl])
            # print(U_iG)
            if self.dyngo:
                # qh = np.linalg.norm(self.lat_vec.T.dot((hkl_G-hkl).T),axis=0)
                qh = Rmat.dot((hkl_G-hkl).T).T.dot(surf_norm)
                sqrtkh = np.sqrt(1+qh/Knorm)
                H[iG,:] += U_iG/(sqrtkg[iG]*sqrtkh)  #so dyngo implementation
            else:
                H[iG,:] += U_iG/(2*self.k0)  #off diagonal terms as potential

        if self.dyngo:
            H/=(2*self.k0)

        H *= 2*np.pi #to get same as felix
        self.H=H
        if show_H:self.show_H()

        if v:print(colors.blue+'...diagonalization...'+colors.black)
        self.gammaj,self.CjG = np.linalg.eigh(self.H) #;print(red+'Ek',lk,black);print(wk)
        self.invCjG = np.linalg.inv(self.CjG)
        self.solved = True

    def _set_structure(self,cif_file):
        pdb_file = ''
        if cif_file[-3:]=='pdb':
            pdb_file=cif_file
            cif_file = ut.pdb2npy(os.path.basename(cif_file[:-4]))

        self.cif_file = cif_file
        self.pdb_file = pdb_file
        self.crys     = ut.import_crys(self.cif_file)
        self.lat_vec0 = np.array(self.crys.lattice_vectors)
        self.lat_vec  = np.array(self.crys.reciprocal_vectors)/(2*np.pi)
        self.pattern  = np.array([np.hstack([a.coords_fractional,a.atomic_number]) for a in self.crys.atoms] )

    def _set_excitation_errors(self,Smax=0.02,hkl=None,felix=False):
        """ get excitation errors for Sg<Smax
        - Smax : maximum excitation error to be included
        - hkl : list of tuple or nbeams x 3 ndarray - beams to be included (for comparison with other programs)
        """
        # print(colors.blue+'... setting excitation error ... '+colors.black)
        K,K0 = self.K,self.k0
        if isinstance(hkl,list) or isinstance(hkl,np.ndarray):
            hkl = np.array(hkl)
            h,k,l = hkl.T
            qx,qy,qz = hkl.dot(self.lat_vec).T
        else:
            (h,k,l),(qx,qy,qz) = self.lattice

        Kx,Ky,Kz = K
        # Sw = K0-np.sqrt((Kx+qx)**2+(Ky+qy)**2+(Kz+qz)**2)
        Sw_full = (K0**2-((Kx+qx)**2+(Ky+qy)**2+(Kz+qz)**2))/(2*K0)
        Sw = Sw_full #;print(Sw)
        # Gz  = -( qx*Kx+qy*Ky+qz*Kz)/K0
        # GzG = Gz -(qx**2+qy**2+qz**2)/(2*K0)
        # print(Sw.shape,Gz.shape,GzG.shape)
        q = np.linalg.norm(np.array([qx,qy,qz]).T,axis=1)
        if felix:
            self.df_G[['qx','qy','qz','q','Sw','Swa']] = np.array([qx,qy,qz,q,Sw,abs(Sw)]).T
            self.df_G['I'] = 0
            self.df_G.loc[str((0,0,0)),'I'] = 1
            return

        # print('ok')
        if Smax:
            idx = abs(Sw)<Smax
            h,k,l = np.array([h[idx],k[idx],l[idx]],dtype=int)
            qx,qy,qz,q,Sw = qx[idx],qy[idx],qz[idx],q[idx],Sw[idx]#,Swa[idx]
            # Gz,GzG = Gz[idx],GzG[idx]
        d = dict(zip(['h','k','l','qx','qy','qz','q','Sw','Swa'],[h,k,l,qx,qy,qz,q,Sw,abs(Sw)]))

        self.Smax = Smax
        self.nbeams = Sw.size
        self.df_G = pd.DataFrame.from_dict(d)
        self.df_G.index = [str(tuple(h)) for h in self.df_G[['h','k','l']].values]
        self.solved = False
        self.df_G['I'] = 0
        self.df_G.loc[str((0,0,0)),'I'] = 1
        # print(self.df_G['I'].shape)


    def _set_Vg(self,felix=0):
        ##### opt was used for testing against FELIX
        if felix:
            q    = self.df_G['q']
            hkl  = self.df_G[['h','k','l']].values
            idx = [i for i,x in enumerate(self.pattern[:,:3]) if all(x<0.99)]
            xi   = self.pattern[idx,:3].T
            # print(xi)
            Za   = np.array(self.pattern[idx,-1],dtype=int)
            fj   = scatf.get_fe(Za,q) #;fj = 1 #testing exp factor alone
            #### Fhkl[nGs] = sum_{natoms} fj [nGs x natoms] * hkl[nGsx3].dot(xi[3,natoms])
            Fhkl = np.sum(fj*np.exp(-2J*np.pi*hkl.dot(xi)),axis=1)
            # print(self.crys,xi)
        else:
            hkl  = self.df_G[['h','k','l']].values
            Fhkl = self.Fhkl.copy()
            V0_idx = np.array([2*self.Nmax]*3)
            Fhkl[tuple(V0_idx)] = 0
            Fhkl = np.array([ Fhkl[tuple(hkl_G+V0_idx)] for hkl_G in hkl])

        self.pre = 1/np.sqrt(1-cst.keV2v(self.keV)**2)
        Vg_G = Fhkl/(self.crys.volume*np.pi)*self.pre*self.eps
        Vg_G = Fhkl
        px,py,e0x = ut.project_beams(K=self.K,qxyz=self.get_G(),e0=[1,0,0],v=1)
        self.e0x = e0x
        self.df_G['px'] = px
        self.df_G['py'] = py
        self.df_G['Vg'] = Vg_G
        self.df_G['Vga'] = abs(Vg_G)
        self.df_G['Swl'] = bloch_util.logM(self.df_G['Sw'])
        self.df_G['L']  = np.ones(Vg_G.shape)
        self._set_zones()
        self.df_G.loc[str((0,0,0)),'Vg'] = 0

    def _set_zones(self):
        hkl     = self.df_G[['h','k','l']].values
        # Khkls   = np.round(hkl.dot(self.Kuvw/np.linalg.norm(self.Kuvw))*100)/100
        Khkls    = np.round(hkl.dot(self.Kuvw0)*100)/100 #integer only in zone axis orientation
        Khkl,ar = np.unique(Khkls,return_inverse=True) #; print(Khkls);print(zones);print(ar)
        zones = np.argsort(Khkl)
        self.df_G['zone'] = zones[ar]

    def _get_central_beam(self):
        return self.get_beam(refl=[str((0,0,0))],index=True)[0]

    def _set_intensities(self):
        """get beam intensities at thickness"""
        id0 = self._get_central_beam()
        gammaj,CjG = self.gammaj,self.CjG
        if self.dyngo:
            S = CjG.dot(np.diag(np.exp(1J*gammaj*self.thick*self.k0/self.Knorm))).dot(self.invCjG)
        else:
            S = CjG.dot(np.diag(np.exp(1J*gammaj*self.thick))).dot(self.invCjG)
        # S = S[:,id0]
        S = S[id0,:]
        self.df_G['S'] = S
        self.df_G['I'] = np.abs(S)**2
        if self.dyngo:
            self.df_G['I']*=self.scale**2

    def _set_I(self,iZ=-1):
        idx=self.get_beam(refl=self.df_G.index)
        self.thick=self.z[iZ]
        self.df_G.I=self.Iz[idx,iZ]

    def _set_kinematic(self):
        Sw,Ug = self.df_G[['Sw','Vg']].values.T
        t,sig = self.thick, self.sig

        #[Ug]=[A-2], [k0]=[A^-1], [t]=[A], [Fhkl]=3[fe]=[A]
        Sg = np.pi/self.k0*Ug*t*np.sinc(Sw*t)
        self.df_G['Sg'] = Sg
        self.df_G['Ig'] = np.abs(Sg)**2

    def _set_beams_vs_thickness(self,thicks=None):
        """ get Scattering matrix as function of thickness for all beams
        - thicks : tuple(ti,tf,step) or list or np.ndarray thicknesses
        """
        print(colors.blue+'... beam vs thickness ...'+colors.black)
        if not type(thicks)==type(None):
            self._set_thicks(thicks)

        id0 = self.get_beam(refl=[(0,0,0)],index=True)[0]#;print(id0)
        gammaj,CjG,invCjG = self.gammaj,self.CjG,self.invCjG

        # import time
        # t0  = time.time()
        #### fast efficient
        # M = np.exp(2J*np.pi*np.outer(gammaj,self.z))*(invCjG[:,id0][:,None])
        M = np.exp(1J*np.outer(gammaj,self.z))*(invCjG[:,id0][:,None])
        St = CjG.dot(M)

        #### slower
        # St = np.zeros((self.df_G.shape[0],self.z.size),dtype=complex)
        # for iT,thick in enumerate(self.z):
        #     S = CjG.dot(np.diag(np.exp(2J*np.pi*gammaj*thick))).dot(invCjG)
        #     # S=CjG.T.dot(np.diag(np.exp(2J*np.pi*gammaj*T))).dot(CjG)
        #     # St[:,iT]=CjG.dot(np.diag(np.exp(2J*np.pi*gammaj*T))).dot(CjG.T)[0,:]
        #     St[:,iT] = S[:,id0]
        # print('done %.2f' %(time.time()-t0))

        self.Sz = St
        self.Iz = np.abs(self.Sz)**2
        Sw,Ug = self.df_G[['Sw','Vg']].values.T
        Sz_kin = np.array([np.pi/self.k0*Ug*t*np.sinc(Sw*t) for t in self.z]).T
        self.Iz_kin = np.abs(Sz_kin)**2

    def _set_thicks(self,thicks):
        if isinstance(thicks,tuple):thicks = np.linspace(*thicks)
        elif isinstance(thicks,float) or isinstance(thicks,int) : thicks=[thicks]
        self.z = np.array(thicks)

    def _get_pkl(self,file=None):
        if not file:file=os.path.join(self.path,self.name+'.pkl')
        return file

    def _get_slice(self,s):
        tle = ''
        if isinstance(s,tuple) or isinstance(s,int) or isinstance(s,str):
            if isinstance(s,int):s = 'l=%d' %s
            if isinstance(s,str):
                i,n = s.split('=')
                tle = 'plane %s, ' %s
                n   = 2*self.Nmax+int(n)
                if   i=='h':s = np.s_[n,:,:]
                elif i=='k':s = np.s_[:,n,:]
                elif i=='l':s = np.s_[:,:,n]
        return s,tle

    ################################################################################
    #### getter
    ################################################################################
    def get_beam(self,
        cond:str='',
        refl:Iterable[tuple or str]=[],
        index:bool=True,
    ) -> Iterable[int]:
        """ select reflections according to a condition or indices.
        Returns strong beams if nothing is specified

        Parameters
        ----------
        cond
            condition on the beams (can be a function see meth:~`bloch_pp.strong_beams` for example)
        list-tuple/str refl
            list of beams either as tuple or str
        index
            if True returns indices as accessible through self.df_G.iloc[idx], otherwise through self.df_G.loc[idx]

        Returns
        -------
        Iterable[int] or Iterable[str]
            index of selected reflections


        .. note::
            reflections will first be seleced by applying cond
        """
        #### ensure self.df_G indices are str
        if not isinstance(self.df_G.index[0],str):
            self.df_G.index = [str(tuple(h)) for h in self.df_G[['h','k','l']].values]

        if not cond and not any(refl):cond={'tol':1e-3}
        if cond:
            if isinstance(cond,dict):
                cond_args=cond.copy()
                if not 'I' in self.df_G.columns:
                    if not 'Iz' in self.__dict__:
                        self._set_beams_vs_thickness(thicks=(0,1000,3))
                    self.df_G.I=self.Iz[:,-1]

                cond=lambda dfG:bloch_util.strong_beams(dfG,**cond_args)
            if isinstance(cond,str):
                refl = list(self.df_G.loc[self.df_G.eval(cond)].index.values)
            else:
                refl=cond(self.df_G)

        if len(list(refl)):
            if not isinstance(refl[0],str):
                refl = [str(tuple(h)) for h in refl]

        #### keep valid reflections only
        refl = [h for h in refl if h in self.df_G.index]

        if index:
            idx = [np.where(self.df_G.index==h)[0][0] for h in refl]
            return idx
        else:
            return refl

    def get_beams_vs_thickness(self,
        dict_opt:bool=False,
        idx:Iterable=slice(0,None,1),
        iZs:Iterable=slice(0,None,1),
    )->np.ndarray:
        """get beams as function of thickness

        Parameters
        ----------
        dict_opt
            returns I(z) as a dictionary
        idx
            beam indices in dataFrame (see meth:~`Bloch.get_beam` for selected beams from their miller indices)
        iZs
            selected thicknesses
        returns
        -------
        dict or np.ndarray
            beams as function of thickness
        """
        if 'Iz' not in self.__dict__:
            self.thicks=(0,1000,1000)
            print('warning : computing Iz for thickness ', self.thicks)
            self._set_beams_vs_thickness()
        Iz = self.Iz[idx,iZs].copy()
        if dict_opt:
            return dict(zip(self.df_G.index[idx], Iz))
        else:
            return Iz


    def get_intensities(self):return self.df_G.I
    def get_hkl(self):return self.df_G[['h','k','l']].values
    def get_kin(self):return self.df_G[['h','k','l','Sw','Vg','Sg','Ig']]
    def get_zones(self):return self.df_G[['h','k','l','zone']].values
    def get_G(self):return self.df_G[['qx','qy','qz']].values
    def get_Xig(self,tol=1e4):
        self.df_G['Xi_g'] = self.k0/abs(self.df_G.Vg)
        xig = self.df_G.loc[self.df_G.Xi_g<tol,['h','k','l','Sw','Vg','Xi_g']]
        print(xig)
        return xig
    def get_Sw(self,Smax=1e-2):
        Smax = min(Smax,self.Smax)
        Smax = self.df_G[['h','k','l','Sw']].loc[self.df_G.Sw<=Smax]
        print(Smax)
        return Smax

    def is_hkl(self,h,k,l):
        """ tests whether beam is part of the simulation
        Parameters
        -----------
            h,k,l
                indices

        Returns
        ---------
            True or False
        """
        return any(self.get_beam(refl=[str((h,k,l))]))

    ################################################################################
    #### display
    ################################################################################
    def show_beams(self,opts='BVSkr',cond='',refl=[],name='',**kwargs):
        """ display beams"""
        # cond = self.get_cond(cond)
        hkl_idx = self.get_beam(cond=cond,refl=refl,index=False)#;print(idx)
        # hkl_idx = self.df_G.index
        # self.df_G.Ig=self.Iz[:,-1]
        self.df_G.loc[str((0,0,0)),'I']=0
        if not name : name=os.path.join(self.path,self.name+'_setup.svg')
        return EDdisp.show_frame(df_bloch=self.df_G,single_mode=False,
            opts=opts,hkl_idx=hkl_idx,name=name,
            **kwargs)

    def show_beams_vs_thickness(self,
        thicks:Optional[Sequence]=None,
        beam_args:dict={},
        refl:Iterable=None,
        iZs:Iterable=slice(0,None,1),
        **kwargs
    ):
        """ display beams vs thickness

        Parameters
        ----------
        thicks
            thickness range (see :meth:~Bloch.set_beams_vs_thickness)
        iZs
            thickness index
        refl
            selected beams
        beam_args
            To pass to get_beam
        kwargs
            to pass to :meth:~EDutil.postprocess.plot_beam_thickness
        """

        if isinstance(thicks,tuple):
            self._set_beams_vs_thickness(thicks)

        ### beam selection
        if isinstance(refl,list):
            if not isinstance(beam_args,dict):beam_args={}
            beam_args['refl']= refl
        if isinstance(beam_args,dict):
            beam_args.update({'index':True})
            idx = self.get_beam(**beam_args)
        else:
            idx = self.get_beam(cond='(Sw<1e-2) & (Vga>0)',index=True)

        z   = self.z[iZs]
        Iz  = self.get_beams_vs_thickness(idx=idx,iZs=iZs,dict_opt=False)
        hkl = self.df_G.index[idx]

        beams=[hkl,z,None,None,Iz]
        return pp.plot_beam_thickness(beams,**kwargs)


    def show_Idyn_vs_Ikin(self,thicks=None,iZs=100,cmap='Spectral',**kwargs):
        """Display correlation plot of dynamical vs kinematical intensities

        Parameters
        -----------
        thicks
            thicknesses
        hkl
            beams to include
        """
        if thicks or not 'Iz' in self.__dict__:self._set_beams_vs_thickness(thicks)
        if isinstance(iZs,int):iZs=slice(0,None,iZs)
        # iB = self.get_Istrong(out=1)[0]#;print(iB,self.Iz[    iB,iZs],)
        # iB = np.argsort(np.sum(self.Iz,axis=1))[-2]
        z      = self.z.copy()[       iZs]
        Iz_dyn = self.Iz.copy()[    :,iZs]#/self.Iz[    iB,iZs]
        Iz_kin = self.Iz_kin.copy()[:,iZs]#/self.Iz_kin[iB,iZs]#;print(Iz_dyn[Iz_dyn>1e-3])
        nzs = z.size
        # print(Iz_kin.shape,Iz_dyn.shape)
        cs = dsp.getCs(cmap,nzs) #; print(len(cs),Iz_dyn.shape,Iz_kin.shape)
        # scat=tuple( [[I_kin,I_dyn,cs[i]] for i,(I_dyn,I_kin) in enumerate(zip(self.Iz[:,::iZ],self.Iz_kin[:,::iZ]))])
        plts=[[np.log10(I_kin),np.log10(I_dyn),[cs[i],'o'],'$z=%d\AA$' %z] for i,(z,I_dyn,I_kin) in enumerate(zip(z,Iz_dyn.T,Iz_kin.T))]
        # plts+=[ [[0,1],[0,1],[(0.5,)*3,'--'],''] ]
        return dsp.stddisp(plts,labs=['$I_{kin}$','$I_{dyn}$'],sargs={'alpha':0.5},**kwargs)

    def show_Fhkl(self,s=None,opts='m',h3D=0,**kwargs):
        """Displays structure factor over grid

        Parameters
        -----------
        opts
            see bloch_util:get_fz
        s
            slice or str('k=0' => Fhkl(k=0)) or int('l==<int>')
        """
        fz,fz_str = bloch_util.get_fz(opts)
        s,s_str = self._get_slice(s)
        tle = 'Structure factor($\AA$), showing %s in %s'  %(fz_str,s_str)
        Fhkl  = self.Fhkl #.copy()
        if isinstance(s,tuple):
            Fhkl   = Fhkl[s]
            nx,ny  = np.array((np.array(Fhkl.shape)-1)/2,dtype=int)
            i,j    = np.meshgrid(np.arange(-nx,nx+1),np.arange(-ny,ny+1))
            fig,ax = dsp.stddisp(scat=[i,j,fz(Fhkl)],title=tle,**kwargs)
        else:
            h,k,l = self.hklF
            fig,ax = dsp.stddisp(scat=[h,k,l,fz(self.Fhkl)],title=tle,rc='3d',**kwargs)
            if h3D:h3d.handler_3d(fig,persp=False)

    def show_H(self,**kwargs):
        dsp.stddisp(im=[np.abs(self.H)],title='abs(H)', pOpt='im')


    ################################################################################
    #### misc
    ################################################################################

    def save(self,file=None,v=1):
        """save this object"""
        file = self._get_pkl(file)
        with open(file,'wb') as out :
            pickle5.dump(self, out, pickle5.HIGHEST_PROTOCOL)
        if v:print(colors.green+"object saved\n"+colors.yellow+file+colors.black)

    def _make_img(self,
            exp:str=None,
            pred:bool=False,
            Imax:int=1,
            Nmax:int=512,
            rot:int=0,
            aperpixel:Optional[float]=None,
            fbroad=None,
            gs3:float=0.1,
            nX:int=1,
            rmax:int=0,
            thick:float=None,
            iz:int=None,
        ):
        '''Make image

        Parameters
        -----------
        exp
            path to dials or Dials (overrides Nmax,aperpixel)
        Imax
            max value for the intensities
        Nmax
            image resolution in pixel
        aperpixel
            reciprocal size for each pixel (aperture per pixel)
        rot
            in plane rotation of the image
        pred
            True to use dials predict (ignores rot)
        fbroad
            broadening function f(r2) (default np.exp(-r2/(gs3/3)**2))
        gs3
            Gaussian broadening factor for each reflections
        nX
            width factor of the Gaussian window (in pixels)
        rmax
            radius for the noise to be added
        show
            Show produced image
        kwargs
            args to be passed to the tiff viewer
        '''
        thick = self.thick
        if thick:self.set_thickness(thick)
        #dials info
        if isinstance(exp,str):exp = pt.Dials(exp)
        if exp:
            aperpixel=exp.aper
            Nmax=exp.nxy
            if not Imax>0 or Imax==1 :
                Imax=exp.Imax*5000
                print(colors.red+'setting Imax to %.1E '% Imax+colors.black)

        hkl = self.df_G.index
        pred = pred and exp
        if pred :
            hkl = [h for h in hkl if h in exp.df_pred.index]
            hkl_lost = np.setdiff1d(self.df_G.index,hkl)
            if any(hkl_lost):
                print(colors.red+"warning : reflections ignored "+colors.black)
                print(self.df_G.loc[hkl_lost].sort_values('I')['I'][-10:])
            pxy = exp.df_pred.loc[hkl,['px','py']]

        # get intensity
        px,py,I = self.df_G.loc[hkl,['px','py','I']].values.T
        Nmax=Nmax//2
        if isinstance(iz,int):
            idb=self.get_beam(refl=hkl)
            I = self.Iz[idb,iz]
            thick = self.z[iz]

        # pixel locations
        if not aperpixel:
            aperpixel = 1.1*max(px.max(),py.max())/Nmax
            print('aperpixel set to %.1E A^-1 ' %(aperpixel))
        dqx,dqy = [aperpixel]*2
        if pred:
            i,j = np.array(pxy[['px','py']].values,dtype=int).T
        else:
            if rot:
                ct,st = np.cos(np.deg2rad(rot)),np.sin(np.deg2rad(rot))
                px,py = ct*px-st*py,st*px+ct*py
            i,j = np.array([np.round(px/dqx),np.round(py/dqy)],dtype=int)+Nmax


        #### kernel (converted to pixel locations)
        if not fbroad:
            fbroad=lambda r2:np.exp(-r2/(gs3/3)**2)
            nx,ny = np.array(np.floor(gs3/np.array([dqx,dqy])),dtype=int)
        else:
            nx,ny=nX,nX
        ix,iy = np.meshgrid(range(-nx,nx+1),range(-ny,ny+1))
        x,y = ix*dqx,iy*dqy
        ## Gaussian broadening
        r2 = (x**2+y**2)
        Pb = fbroad(r2)#;dsp.stddisp(im=[x,y,Pb],pOpt='im')
        im0 = np.zeros((2*Nmax,)*2)
        for i0,j0,I0 in zip(i,j,I):
            i0x,j0y = i0+ix,j0+iy
            idx     = (i0x>=0) & (j0y>=0)  & (i0x<2*Nmax) & (j0y<2*Nmax)
            i0x,j0y = i0+ix[idx],j0+iy[idx]
            im0[i0x,j0y] += Pb[idx]/Pb[idx].sum()*I0
            # im0[i0,j0]=I0

        #### noise
        if rmax:
            h,k = np.meshgrid(range(-Nmax,Nmax),range(-Nmax,Nmax))
            r = np.sqrt(h**2+k**2);r[r==0]=1
            im0 += rmax*np.random.rand(*im0.shape)#/(rmax*r)

        return im0*Imax

    def convert2img(self,filename,template=None,**kwargs):
        im0 = self._make_img(**kwargs)#Nmax,fbroad,gs3,nX,rmax,thick,iz,rot)
        # print(im0.mean())
        fmt = template.split('.')[-1]
        if not filename:
            filename = os.path.join(self.figpath,self.name+'_%dA.%s' %(self.thick,fmt))
        if template:
            out = check_output("cp %s %s" %(template,filename),shell=True).decode()
            if out:print(out)
        bloch_util.imwrite(filename,im0)



    def convert2tiff(self,tiff_file:str=None,
        Imax:int=3e4,
        Nmax:int=512,aperpixel:Optional[float]=None,
        fbroad=None,
        gs3:float=0.1,
        nX:int=1,
        rmax:int=0,
        thick:float=None,
        iz:int=None,
        rot:int=0,
        show:bool=False,
        tif_writer_args:dict={},
        **kwargs,
    ):
        """Write intensities to a tiff file

        Parameters
        -----------
        tiff_file
            name of the file (default name if not provided)
        thick
            Thickness of the simu
        Imax
            max value for the intensities
        Nmax
            image resolution in pixel
        aperpixel
            reciprocal size for each pixel (aperture per pixel)
        fbroad
            broadening function f(r2) (default np.exp(-r2/(gs3/3)**2))
        gs3
            Gaussian broadening factor for each reflections
        nX
            width factor of the Gaussian window
        rmax
            radius for the noise to be added
        show
            Show produced image
        kwargs
            args to be passed to the tiff viewer


        .. note ::
            The reciprocal space resolution is aperpixel*Nmax. As a result some of teh
            high resolution reflections might not appear if aperpixel or Nmax are too small.
            If aperpixel is not specified, it is automatically so it contains the image
            will contain all reflections.
        """
        im0 = _make_img(Nmax,fbroad,gs3,nX,rmax,thick,iz,rot)

        if not tiff_file:
            tiff_file = os.path.join(self.path,self.name+'_%dA' %thick+'.tiff')
        I = np.array(im0*Imax,dtype='uint16')

        # ix,iy = np.meshgrid(range(2*Nmax),range(2*Nmax))
        # dsp.stddisp(im=[ix,iy,I],plots=[j,i,'bo'],xylims=[0,512,0,512],
        #     cmap='gray',caxis=[0,10],imOpt='tX',pargs={'fillstyle':'none'})

        tifffile.imwrite(tiff_file,I.T,**tif_writer_args)
        print(colors.yellow+tiff_file+colors.green+' saved'+colors.black)
        if show:
            # v=viewers.Base_Viewer(self.path,frame=1,thick=self.thick,**kwargs)
            return ut.show_tiff(tiff_file,**kwargs)
            # return v
        else:
            return tiff_file


    ###################################################################################
    #### convergence stuff
    ###################################################################################
    def convergence_test(self,
        Nmax:Iterable[int],
        Smax:Iterable[float],
        z=(0,100,100),
        hkl=None,
        opts:str='Ss'
    ):
        """creates a convergence test for a range of Nmax and Smax
        and displays it as funtion of thickness for beams hkl

        Parameters
        ----------
        opt:  S(solve) s(save) p(plot)
        """
        solve,save,show = 'S' in opts,'s' in opts,'p' in opts
        Nmax,Smax=[a.flatten() for a in np.meshgrid(Nmax,Smax)]
        z=np.array(z)
        df=pd.DataFrame(columns=['Nmax','Smax','nbeams','Iz'])
        # simulate
        self.update_Nmax(Nmax[0])
        self._set_excitation_errors(Smax=Smax[0])
        hkl0 = self.df_G.index
        if solve:
            for i,(Nm,Sm) in enumerate(zip(Nmax,Smax)):
                 self.solve(Nmax=Nm,Smax=Sm,thicks=z,opts='v0')
                 idx='(%d,%.3f)'%(Nm,Sm)
                 idx_b = self.get_beam(refl=hkl0)
                 df.loc[idx,['Nmax','Smax','nbeams']]=[Nm,Sm,self.nbeams]
                 df.loc[idx,'Iz']=self.Iz[idx_b,:].copy()
            self.df=df
            if save:
                df.to_pickle(self.path+self.name+'_cv.pkl')

        if show:return self.show_convergence(hkl)
        # else:
        #     for i,(Nm,Sm) in enumerate(zip(Nmax,Smax)):
        #         self.update_Nmax(Nm)
        #         self._set_excitation_errors(Smax=Sm)
        #         idx='(%d,%.3f)'%(Nm,Sm)
        #         df.loc[idx,['Nmax','Smax','nbeams']]=[Nm,Sm,self.nbeams]
        #     print(df)

    def show_convergence(self,hkl,xlab='Smax',**kwargs):
        #beam selection
        valid_hkl=0
        # if isinstance(hkl,list):valid_hkl=isinstance(hkl[0],str)
        self.update_Nmax(int(self.df.Nmax.min()))
        self._set_excitation_errors(Smax=self.df.Smax.min())
        # if not valid_hkl:
        #     hkl = self.get_beam(cond='Sw<1e-2' ,opt=0)[0:2]
        #     hkl += [self.get_beam(cond='Sw>=1e-2',opt=0)[0]]
        #     print('default hkl : ',hkl)
        # Icols=hkl
        # self.df[Icols]=None
        x = self.df[xlab]
        idx_s = self.get_beam(refl=hkl)
        #Iz.shape:nbeams x nsimus
        Iz = np.array([r.Iz[idx_s,-1] for i,r in self.df.iterrows()]).T
        cs=dsp.getCs('Spectral',len(hkl))
        #plot
        plts=[ [x,I,[c,'-o'],h]  for i,h,c,I in zip(idx_s,hkl,cs,Iz)]

        # ## thickness dependence
        # cs=dsp.getCs('RdYlGn',self.df.shape[0])
        # ms = dsp.markers[:len(hkl)]
        # for idx,m in zip(idx_s,ms):
        #     plts+=[[self.z,r.Iz[idx],[c,m+'-' ],
        #          ] for c,(l,r) in zip(cs,self.df.iterrows())]
        # legElt={'$%s$' %h:'k-'+ms[i] for i,h in enumerate(hkl)}
        # legElt.update({'$%s$' %l:[c,'-'] for c,l in zip(cs,self.df.index)})
        # return dsp.stddisp(plts,labs=['$z$','$I$'],legElt=legElt,lw=2)
        return dsp.stddisp(plts,labs=['$%s$' %xlab,'$I$'],lw=2,**kwargs)


# from wallpp import wallpaper as wallpp  ;imp.reload(wallpp)
# from multislice import mupy_utils as mut;imp.reload(mut)
# class Bloch2D(Bloch):
#     def _set_structure(self,file):
#         '''wallpaper file (see import_wallpp)'''
#         self.file = file
#         if not crys:crys=file
#         self.crys = mut.import_wallpp(crys)
#         # self.crys = wallpp.Wallpaper(**crys)
#         self.lat_vec0 = self.crys.lattice_vectors
#         self.lat_vec  = self.crys.reciprocal_vectors#/(2*np.pi)
#         self.pattern  = self.crys.pattern_fract
#
#     def convert2tiff(self):
#         printf('not relevant in 2D')
#     def show_Fhk(self,opts='m',**kwargs):
#         """Displays structure factor over grid
#
#         Parameters
#         -----------
#         opts
#             see bloch_util:get_fz
#         """
#         fz,fz_str = bloch_util.get_fz(opts)
#         Fhk   = self.Fhk
#         nx,ny  = np.array((np.array(Fhk.shape)-1)/2,dtype=int)
#         i,j    = np.meshgrid(np.arange(-nx,nx+1),np.arange(-ny,ny+1))
#         return dsp.stddisp(scat=[i,j,fz(Fhk)],title=tle,**kwargs)
#
#     def update_Nmax(self,Nmax:int):
#         """
#         Update resolution/max order
#
#         Parameters
#         ----------
#         Nmax
#             maximum h,k order
#         """
#         if type(Nmax) in [int,np.int64] :
#             if not Nmax==self.Nmax:
#                 self.Nmax=Nmax
#                 (h,k),(qx,qz) = mut.get_lattice2D(self.lat_vec,self.Nmax)
#                 self.lattice = [(h,k),(qx,qz)]
#                 self.hkF,self.Fhk = sf.structure_factor2D(self.pattern, 2*np.pi*self.lat_vec,hkMax=2*self.Nmax,eps=self.eps)
#
#
#     def _get_central_beam():
#         return self.get_beam(refl=[str(0,0)],index=True)[0]
#     def _set_zones(self):return
#
#     def _set_excitation_errors(self,Smax=0.02):
#         ''' get excitation errors for Sg<Smax
#         - Smax : maximum excitation error to be included
#         '''
#         K,K0 = self.K,self.k0
#         (h,k),(qx,qy) = self.lattice
#
#         Kx,Ky = K
#         Sw = np.abs(np.sqrt((Kx-qx)**2+(Ky-qy)**2) - K0)
#         if Smax:
#             idx = Sw<Smax
#             h,k = np.array([h[idx],k[idx]],dtype=int)
#             qx,qy,Sw = qx[idx],qy[idx],Sw[idx]
#         d = dict(zip(['h','k','qx','qy','Sw'],[h,k,qx,qy,Sw]))
#
#         self.Smax   = Smax
#         self.nbeams = Sw.size
#         self.df_G   = pd.DataFrame.from_dict(d)
#         self.solved = False
#
#     def _set_Vg(self):
#         #in 2D Fhk is just the Fourier transform of fv
#         vg = self.Fhk.copy()
#         #It needs to be converted into a scattering amplitude [A]
#         #Fg[no dim] = vg[kVA^2], sig=[1/kVA], k0[A-1]
#         Fg = vg*self.sig*self.k0/np.pi
#         # Vg[A^-2]
#         Vg_G =  Fg/self.crys.area
#
#         #self.pre = 1/np.sqrt(1-cst.keV2v(self.keV)**2)
#         # Vg_G = Fhkl/(self.crys.volume*np.pi)*self.pre*self.eps
#         # px,py,e0x = ut.project_beams(K=self.K,qxyz=self.get_G(),e0=[1,0,0],v=1)
#         # self.e0x = e0x
#         # self.df_G['px'] = px
#         # self.df_G['py'] = py
#         self.set_beam_positions()
#         self.df_G['Vg'] = Vg_G
#         self.df_G['Vga'] = abs(Vg_G)
#         self.df_G['Swl'] = bloch_util.logM(self.df_G['Sw'])
#         self.df_G['L']  = np.ones(Vg_G.shape)
#         self._set_zones()
#         self.df_G.loc[str((0,0)),'Vg'] = 0
#
#     def set_beam_positions(self):
#         self.df_G['px'] = mut.project_beams2D(K=self.K,qxy=self.get_G())



import importlib as imp
import json,tifffile,os,glob,pickle5,subprocess,crystals
import numpy as np,pandas as pd
from typing import TYPE_CHECKING, Dict, Iterable, Optional, Sequence, Union
from crystals import Crystal
from utils import displayStandards as dsp   #;imp.reload(dsp)
from utils import glob_colors as colors     #;imp.reload(colors)
# from utils import handler3D as h3D          #;imp.reload(h3D)
# from . import rotate_exp as exp ;imp.reload(exp)
import gemmi

def sweep_var(Simu:object,
        params:Sequence[str],vals:Sequence[Sequence],
        tag:str='',path:str='',
        **kwargs,
    ):
    """runs a set of similar simulations Simu with a varying parameter

    Parameters
    ----------
    Simu
        Simulator object
    params
        list of parameters to be changed (can be a single string)
    vals
        values to assign to the parameters (nsimus x nparams)
    tag
        Name prefix of simulations
    path
        path to the simulations folder
    kwargs
        arguments passed to Simu

    Returns
    -------
        pd.DataFrame
            info about the set of simulations with params pkl
    """
    if not os.path.exists(path):
        print('creating directory:',path)
        print(subprocess.check_output('mkdir -p %s' %path,shell=True).decode())

    if isinstance(params,str):
        params=[params]
        vals=[[v] for v in vals]

    if not isinstance(vals,list):
        vals=list(vals)
    if not isinstance(vals[0],list):
        vals = [list(v) for v in vals]

    if not len(params)==len(vals[0]):
        raise Exception('len(params)=%d but it must equal len(vals[0])=%d' %(len(params),len(vals[0])))

    cols = params+['pkl']
    df = pd.DataFrame(columns=cols)#+pp.info_cols)

    nsimus = len(vals)
    pad = int(np.ceil(np.log10(nsimus)))
    for i,val in enumerate(vals):
        kwargs.update(dict(zip(params,val)))
        name = '%s_%s%s' %('-'.join(params),tag,str(i).zfill(pad))
        sim_obj = Simu(path=path,name=name,**kwargs)
        df.loc[name,cols] = ''
        # df.loc[name,cols] = list(val)+[sim_obj._get_pkl()]
        df.loc[name,params] = list(val)
        df.loc[name,'pkl']  = sim_obj._get_pkl()
    df_file = os.path.join(path,'df_%s.pkl' %tag)
    df.to_pickle(df_file)
    print(colors.green+'Dataframe saved : '+colors.yellow+df_file+colors.black)
    return df

def get_pkl(file=None,path='',name='unkown.pkl'):
    if not file:
        file=os.path.join(path,name+'.pkl')
    return file

def save(obj,file=None,path='',name='unkown.pkl'):
    """save an object"""
    file = get_pkl(file,path,name)
    with open(file,'wb') as out :
        pickle5.dump(obj, out, pickle5.HIGHEST_PROTOCOL)
    print(colors.green+"object saved\n"+colors.yellow+file+colors.black)
def save_pkl(obj,file):
    save(obj,file=file)

def load_pkl(file):
    """load an object"""
    with open(file,'rb') as f : obj = pickle5.load(f)
    return obj

def rot(a,axis='x',deg=True):
    if deg:a = np.deg2rad(a)
    c,s = np.cos(a),np.sin(a)
    if   axis=='x':R = np.array([[1,0,0],[0,c,s],[0,-s,c]])
    elif axis=='y':R = np.array([[c,0,s],[0,1,0],[-s,0,c]])
    elif axis=='z':R = np.array([[c,s,0],[-s,c,0],[0,0,1]])
    return R

def rotation_matrix(k,a,deg=True):
    '''Rotation matrix from a vector'''
    #https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    if deg:a = np.deg2rad(a)
    kx,ky,kz = k
    K = np.array([[0,-kz,ky],[kz,0,-kx],[-ky,kx,0]])
    R = np.eye(3) + np.sin(a)*K + (1-np.cos(a))*K.dot(K)

    return R
#############################################################################
#### orientation vectors
#############################################################################
def get_uvw_cont(u0:Sequence,u1:Sequence,nframes:int=2,show:bool=False,**kwargs):
    """get a list of orientation vectors in a continuously rotated fashion between u0 and u1.

    Parameters
    ----------
    u0
        initial vector
    u1
        final vector
    nframes
        number of vectors inbetween
    show
        show the vectors
    kwargs
        :func:~utilities.show_uvw
    """
    u0,u1=np.array(u0,dtype=float),np.array(u1,dtype=float)
    u0/=np.linalg.norm(u0)
    u1/=np.linalg.norm(u1)
    e1 = u1 - u0.dot(u1)*u0
    e1/=np.linalg.norm(e1)

    omega_max = np.arccos(u1.dot(u0))
    omega_r = np.linspace(0,omega_max,nframes)

    ct,st = np.cos(omega_r),np.sin(omega_r)
    u0e1  = np.hstack([np.array(u0)[:,None],np.array(e1)[:,None]])
    xy    = np.vstack([ct,st])
    uvw   = u0e1.dot(xy).T

    if show:
        if not 'eij' in kwargs.keys():kwargs['eij']={}
        kwargs['eij'].update(dict(zip(['u_0','u_1'],[u0,u1])))
        show_uvw(uvw,**kwargs)
    return uvw

def get_uvw_rock(e0,e1=[0,1],deg:float=1,dt:Optional[float]=None,npts:int=20,**kwargs):
    """create orientation vectors around u0 to simulate a rocking curve

    Parameters
    ----------
    e0
        main beam orientation
    e1
        rocking curve direction components in spherical basis ( u1[x,y,z] = u1[0]*e_theta+u1[1]*e_phi)
    deg
        angular range in degrees
    dt
        angular step in degrees
    npts
        number of points (angular spread npts*deg)
    kwargs
        passed to :func:~utilities.get_uvw_cont
    """
    e0=np.array(e0,dtype=float)/np.linalg.norm(e0)

    theta,phi = np.deg2rad(theta_phi_from_u(e0))    #euler angles
    ct,st,cp,sp = np.cos(theta),np.sin(theta),np.cos(phi),np.sin(phi)
    e_theta = np.array([ct*cp,ct*st,-st])
    e_phi = np.array([-sp,cp,0])
    e1  = e1[0]*e_theta+e1[1]*e_phi
    e1 /= np.linalg.norm(e1)

    if dt:
        deg = dt*npts
        print('angular spread=%.1f degrees' %deg)
    deg_r = np.deg2rad(deg)
    u0 = e0-e1*deg_r
    u1 = e0+e1*deg_r

    if not 'eij' in kwargs.keys():kwargs['eij']={}
    kwargs['eij'].update(dict(zip(['e_0'],[e0])))
    uvw = get_uvw_cont(u0,u1,nframes=npts,**kwargs)
    return uvw

def get_uvw_CBED(u0:Sequence,deg:float,npts:int,
    cart:bool=True,show:bool=False,**kwargs
    ):
    """get orientation vectors in a CBED pattern

    Parameters
    ----------
    u0
        main orientation
    deg
        angular range in degrees
    npts
        number of points around u0 in each direction
    cart
        compute using cartesian transform
    show
        show uvw
    kwargs
        passed to :func:~utilities.show_uvw
    """
    u0 = np.array(u0,dtype=float)
    e0 = u0/np.linalg.norm(u0)
    e1 = np.cross(e0,[0,1,0])
    e2 = np.cross(e0,e1)
    eij = np.vstack([e0,e1,e2])
    dt = np.deg2rad(deg)

    if cart:
        dx = np.tan(dt)
        x0 = np.linspace(-dx,dx,npts)
        x,y = np.meshgrid(x0,x0)
        x,y = x.flatten(),y.flatten()
        uvw = np.vstack([x,y]).T.dot(np.vstack([e1,e2]))+e0
        # uvw += e0
        uvw = (uvw.T/np.linalg.norm(uvw,axis=1)).T
    else:
        if isinstance(npts,int):npts=[npts]*2
        ds = np.linspace(0,dt,npts[0])
        theta,phi = np.meshgrid(ds,np.linspace(0,2*np.pi,npts[1]))
        theta,phi = theta.flatten(),phi.flatten()
        ct,st,cp,sp = np.cos(theta),np.sin(theta),np.cos(phi),np.sin(phi)
        uvw0 = np.vstack([ct,st*cp,st*sp]).T
        uvw = uvw0.dot(eij)
        # uvw /=np.linalg.norm(uvw,axis=1)

    if show:show_uvw(uvw,eij=dict(zip(['e_0','e_1','e_2'],eij)),**kwargs)
    return uvw

def show_uvw(uvw:Iterable[Sequence],ax=None,
        eij:dict={'e_z':[0,0,1]},h3d:bool=False,**kwargs
    ):
    """show orientation vectors

    Parameters
    ----------
    uvw
        Array of orientations
    eij
        additional vectors to plot
    h3d
        use vector handlers
    """
    ez =  np.cross(uvw[0],uvw[-1])
    cs = dsp.getCs('viridis',len(list(eij.keys())) )

    plts = [[[0,u[0]],[0,u[1]],[0,u[2]],'b'] for u in uvw ]
    plts+= [[[0,ei[0]],[0,ei[1]],[0,ei[2]] , c ,'$%s$' %k] for c,(k,ei) in zip(cs,eij.items())]
    fig,ax = dsp.stddisp(plots=plts,labs=['x','y','z'],rc='3d',lw=4,xylims=1,**kwargs)
    if h3d:h3d = h3D.handler_3d(fig,persp=False,xm0=1)
    return fig,ax

def u_from_theta_phi(theta,phi):
    theta,phi = np.deg2rad([theta,phi])
    ct,st,cp,sp = np.cos(theta),np.sin(theta),np.cos(phi),np.sin(phi)
    u = [st*cp,st*sp,ct]
    return u

def theta_phi_from_u(u):
    u = u/np.linalg.norm(u)
    x,y,z  = u
    theta = np.rad2deg(np.arccos(z))
    phi   = np.rad2deg(np.arctan2(y,x))
    return theta,phi

def get_uvw_from_theta_phi(theta,phi,e1=[0,0,1],**kwargs):
    u0 = np.array(u_from_theta_phi(theta,phi))
    u1 = np.cross(e1,u0);u1/=np.linalg.norm(u1)
    uvw = get_uvw_cont(u0,u1,**kwargs)
    return uvw

def uvw_add_points(uvw0,npts=1,plot=0,h3d=1,**kwargs):
    ntot = (uvw0.shape[0]-1)*(npts+1) + 1
    uvw = np.zeros((ntot,3))
    for i in range(uvw0.shape[0]-1):
        e0,e1 = uvw0[i:i+2,:]
        # e0,u1 = uvw0[i:i+2,:]
        # e1 = u1 - e0.dot(u1)*e0
        # e1/=np.linalg.norm(e1)
        # alpha = np.rad2deg(np.arccos(u1.dot(e0)))#;print(alpha)
        uvw_i = get_uvw_cont(e0,e1,nframes=npts+2)#omega=np.linspace(0,alpha,npts+2))
        idx = i*(npts+1)
        uvw[idx:idx+npts+1,:] = uvw_i[:-1,:]
    uvw[-1,:] = uvw_i[-1,:]

    if plot:
        plts  = [[[0,u[0]],[0,u[1]],[0,u[2]],'r'] for u in uvw ]
        fig,ax = dsp.stddisp(plots=plts,labs=['x','y','z'],rc='3d',lw=4,opt='')
        plts = [[[0,u[0]],[0,u[1]],[0,u[2]],'b'] for u in uvw0 ]
        dsp.stddisp(ax=ax,plots=plts,labs=['x','y','z'],rc='3d',lw=2,**kwargs)
        # if h3d:h3d = h3D.handler_3d(fig,persp=False,xm0=1)
        return fig,ax
    return uvw



######################################################
#### misc
######################################################
def _find_files(path,f_type):
    files = glob.glob(path+'/*.%s' %f_type)
    if len(files)==1:
        file=files[0]
        return file
    else:
        msg ='''
    only 1 %s file should be found in path folder :
     %s
    but %d were found :
     %s''' %(f_type,os.path.realpath(path),len(files),str(files))
        raise Exception(msg)

def find_cif_file(path,cif_file=None):
    """find cif file in path"""
    if not cif_file:
        cif_file = _find_files(path,'cif')
    return cif_file

def pdb2npy(pdb,npy_file=''):
    crys=crystals.Crystal.from_pdb(pdb)
    Zxyz=np.array([
        np.array([a.atomic_number,a.coords_fractional%1],dtype=object)
            for a in crys],dtype=object)
    if not npy_file: npy_file=pdb+'.npy'
    np.save(npy_file,np.array([Zxyz ,crys.lattice_vectors],dtype=object))
    print(colors.green+'file saved : '+colors.yellow+npy_file+colors.black)
    return npy_file

def import_npy(npy_file):
    (Zxyz,lat) = np.load(npy_file,allow_pickle=True)
    crys=crystals.Crystal(
        crystals.AtomicStructure([
            crystals.Atom(element=Z,coords=xyz) for Z,xyz in Zxyz]),
        lat)
    return crys

def crys2felix(crys,opt='wr',out=None):
    # crys = Crystal.from_cif(cif_file)
    cif = """_chemical_formula_sum '%s'
loop_
_cell_length_a %.5f
_cell_length_b %.5f
_cell_length_c %.5f
_cell_angle_alpha %.3f
_cell_angle_beta %.3f
_cell_angle_gamma %.3f
_symmetry_Int_Tables_number %d
_symmetry_space_group_name_Hall '%s'
loop_
_atom_site_label
_atom_site_type_symbol
_atom_site_fract_x
_atom_site_fract_y
_atom_site_fract_z
_atom_site_B_iso_or_equiv
_atom_site_occupancy
""" %(
        crys.chemical_formula,*crys.lattice_parameters,
        crys.international_number,crys.hall_symbol)
    cif += '\n'.join([
        "%s %s %.5f %.5f %.5f 0.1 1" %("%s%d" %(a.element,i), a.element,*a.coords_fractional)
            for i,a in enumerate(crys.atoms) if all(a.coords_fractional<0.99)
        ])

    if not out:out='felix.cif'
    if 'w' in opt:
        with open(out,'w') as f : f.write(cif)
        print(colors.yellow+out+colors.green+" successfully created."+colors.black)
    if 'r' in opt:
        if 'w' in opt:
            print(colors.blue+"Content of "+ colors.yellow +out+colors.black)
            with open(out,'r') as f : print(''.join(f.readlines()))
        else:
            print(cif)

def gemmi_sf(pdb_file:str='',dmin=2):
    st = gemmi.read_structure(pdb_file)
    dc = gemmi.DensityCalculatorX()
    dc.d_min = dmin
    dc.addends.subtract_z()
    dc.set_grid_cell_and_spacegroup(st)
    dc.set_refmac_compatible_blur(st[0])
    dc.put_model_density_on_grid(st[0])
    grid = gemmi.transform_map_to_f_phi(dc.grid); print('sf : ',grid.shape)
    Nmax = (np.array(grid.shape)//2-1).min()
    Fhkl = np.zeros((2*Nmax+1,)*3,dtype=complex)
    hklF = np.meshgrid(*[np.arange(-Nmax,Nmax+1)]*3)
    hkls = np.array([i.flatten() for i in hklF]).T
    idxs = np.array([i.flatten() for i in np.meshgrid(*[np.arange(2*Nmax+1)]*3)]).T
    #TODO : multiple size grid
    # nu2,nv2,nw2 = np.array(grid.shape)//2-1
    # Fhkl = np.zeros((2*nu2+1,2*nv2+1,2*nw2+1),dtype=complex)
    # hkls=np.array([i.flatten() for i in np.meshgrid(range(-nu2,nu2+1),range(-nv2,nv2+1),range(-nw2,nw2+1))]).T
    # idxs=np.array([i.flatten() for i in np.meshgrid(range(2*nu2+1),range(2*nv2+1),range(2*nw2+1))]).T
    # hkls = pd.DataFrame()
    print(colors.blue+'...filling...'+colors.black)
    for idx,hkl in zip(idxs,hkls):
        Fhkl[tuple(idx)] = dc.mott_bethe_factor(hkl) * grid.get_value(*hkl)
    return hklF,Fhkl

def import_crys(file:str=''):
    """import a Crystal

    Parameters
    ----------
    file
        can be a cif file or a builtin
    """
    builtins=list(Crystal.builtins)

    if file.split('.')[-1]=='cif':
        crys = Crystal.from_cif(file)
    elif file.split('.')[-1]=='npy':
        crys = import_npy(file)
    elif sum(np.array(list(Crystal.builtins))==file):
        crys = Crystal.from_database(file)
    else:
        if file:
            msg = 'cannot import %s\n' %file
            msg+= 'Provide a .cif file or use a builtin structure \n%s' %builtins.__str__()
            raise Exception(msg)
        else:
            print('available builtins:\n')
            print(builtins.__str__())
            return
    print(colors.green+'imported file : '+colors.yellow+file+colors.black)
    return crys

def convert2tiff(tiff_file,im0,n0=0,rot=0,n=256,Imax=5e4):
    '''
    used to rotate a multislice image to a standard pets tiff file image
    '''
    alpha = np.deg2rad(rot)
    ct,st = np.cos(alpha),np.sin(alpha)
    u_n   = np.arange(-n,n+1)
    hi,ki = np.meshgrid(u_n,u_n)
    Hi,Ki = ct*hi+st*ki,st*hi-ct*ki
    H,K = np.array(np.round(Hi)+n0,dtype=int),np.array(np.round(Ki)+n0,dtype=int)
    # dsp.stddisp(im=[I[H,K]],caxis=[0,50],pOpt='im',cmap='gray')
    I = np.array(im0*Imax,dtype='uint16')
    tifffile.imwrite(tiff_file,I[H,K]) #,np.flipud(I))
    print(colors.yellow+tiff_file+colors.green+' saved'+colors.black)




################################################################################
#### Reciprocal space related
################################################################################
def get_reciprocal(abc):
    a1,a2,a3 = abc
    b1 = np.cross(a2,a3)/(a1.dot(np.cross(a2,a3)))
    b2 = np.cross(a3,a1)/(a2.dot(np.cross(a3,a1)))
    b3 = np.cross(a1,a2)/(a3.dot(np.cross(a1,a2)))
    abc_star = np.vstack([b1,b2,b3])#*(2*np.pi)
    return abc_star

def get_lattice(lat_vec,Nmax=5):
    '''Get a lattice up to Nmax points :
    - lat_vec : reciprocal lattice vectors
    - Nmax : max order '''
    a1,b1,c1 = lat_vec
    N = np.arange(-Nmax,Nmax+1)
    h,k,l = np.meshgrid(N,N,N)
    h,k,l = h.flatten(),k.flatten(),l.flatten()
    qx = h*a1[0]+k*b1[0]+l*c1[0]
    qy = h*a1[1]+k*b1[1]+l*c1[1]
    qz = h*a1[2]+k*b1[2]+l*c1[2]
    return (h,k,l),(qx,qy,qz)
def get_lattice2D(lat_vec,Nmax=5):
    '''Get a lattice up to Nmax points :
    - lat_vec : reciprocal lattice vectors
    - Nmax : max order '''
    a1,b1 = lat_vec
    N = np.arange(-Nmax,Nmax+1)
    h,k = np.meshgrid(N,N)
    h,k = h.flatten(),k.flatten()
    qx = h*a1[0]+k*b1[0]
    qz = h*a1[1]+k*b1[1]
    return (h,k),(qx,qz)

def get_ewald(K,nts=100,nps=200):
    ''' Get ewald sphere coordinates
    - K : reciprocal space beam vector
    '''
    Kx,Ky,Kz = K

    theta,phi = np.linspace(0,np.pi,nts),np.linspace(0,2*np.pi,nps)
    x = Kx+self.K0*np.outer(np.sin(theta),np.cos(phi))
    y = Ky+self.K0*np.outer(np.sin(theta),np.sin(phi))
    z = Kz+self.K0*np.outer(np.cos(theta),np.ones(phi.shape))
    return x,y,z

def get_excitation_errors(K,lat_vec,Nmax=5,Smax=0.02):
    ''' get excitation errors for lattice lat_vec and beam K
    - K : reciprocal space beam vector
    - lat_vec : reciprocal lattice vectors
    - Nmax : max order of reflections(resolution)
    - Smax : maximum excitation error to be included
    '''
    K0 = np.linalg.norm(K)
    Kx,Ky,Kz = K

    (h,k,l),(qx,qy,qz) = get_lattice(lat_vec,Nmax)

    # Sw = np.abs(np.sqrt((Kx-qx)**2+(Ky-qy)**2+(Kz-qz)**2) - K0)
    Sw = np.abs(np.sqrt((Kx+qx)**2+(Ky+qy)**2+(Kz+qz)**2) - K0)
    if Smax:
        idx = Sw<Smax
        h,k,l = np.array([h[idx],k[idx],l[idx]],dtype=int)
        qx,qy,qz,Sw = qx[idx],qy[idx],qz[idx],Sw[idx]
    d = dict(zip(['h','k','l','qx','qy','qz','Sw'],[h,k,l,qx,qy,qz,Sw]))
    df = pd.DataFrame.from_dict(d)
    df.index=[str(tuple(h)) for h in df[['h','k','l']].values]
    return df

def get_structure_factor(cif_file,dfout=0,**sf_args):
    '''computes structure factor 3D
    - sf_args : see (structure_factor3D)
    returns :
    - (qx,qy,qz),Fhkl
    '''
    crys = import_crys(cif_file)
    pattern = np.array([np.hstack([a.coords_fractional,a.atomic_number]) for a in crys.atoms] )
    lat_vec = np.array(crys.reciprocal_vectors)
    (h,k,l),Fhkl = structure_factor3D(pattern, lat_vec, **sf_args)
    qx = h/crys.lattice_parameters[0]
    qy = k/crys.lattice_parameters[1]
    qz = l/crys.lattice_parameters[2]
    if dfout:
        data = [h.flatten(),k.flatten(),l.flatten(),qx.flatten(),qy.flatten(),qz.flatten(),Fhkl.flatten()]
        d = dict(zip(['h','k','l','qx','qy','qz','Fhkl'],data))
        return pd.DataFrame.from_dict(d)
    else:
        return (qx,qy,qz),Fhkl

def get_excited_beams(cif_file,K,lat_vec,Nmax,Smax):
    df_Sw   = get_excitation_errors(K,lat_vec,Nmax,Smax)
    df_Fhkl = get_structure_factor(cif_file,dfout=1,hklMax=Nmax+1)

    hkl0 = df_Sw[['h','k','l']].values
    hklF = df_Fhkl[['h','k','l']].values
    ridx = []
    for r in hkl0:
        idx = np.where(np.linalg.norm(r-hklF,axis=1)==0)[0]
        if idx.size:
            ridx+=[idx[0]]

    Fhkl_0 = df_Fhkl.iloc[ridx]['Fhkl'].values #;print(df_Fhkl.iloc[ridx]);#print(ridx);print(df_Fhkl.shape)
    df_Sw['Fhkl'] = Fhkl_0
    return df_Sw

def get_kinematic_intensities(cif_file,K,thick,Nmax=5,Smax=None):
    crys    = import_crys(cif_file)
    lat_vec = np.array(crys.reciprocal_vectors)/(2*np.pi)
    df_Sw   = get_excited_beams(cif_file,K,lat_vec,Nmax,Smax)

    #kinematic intensities
    K0  = np.linalg.norm(K)
    sig = cst.keV2sigma(cst.lam2keV(1/K0))      #;print(sig)
    Sw  = df_Sw.Sw.values
    gs = (sig*np.pi*thick*Sw)*np.sinc(thick*Sw)

    Fhkl_0 = df_Sw['Fhkl'].values
    I  = (np.abs(Fhkl_0)*gs)**2
    df_Sw['I'] = I
    return df_Sw

def project_beams(K,qxyz,e0=[1,0,0],v=0):
    '''pass from reciprocal cartesian coordinates to panel coordinates
    then uses the `aper` parameter to get values in pixels
    e0:rotation axis in the panel frame
    Note:
        In the images frame, the beam in perpendicular to the images(z axis)
        and the rotation axis is usually along the x axis of the images
    '''
    #build the frames basis
    e3 = K/np.linalg.norm(K)
    if not np.linalg.norm(e3-e0):
        if not np.linalg.norm(e3-np.array([1,0,0])):
            e0=[0,1,0]
        else:
            e0=[1,0,0]

    e0 = np.array(e0)/np.linalg.norm(e0)
    e2 = np.cross(e3,e0)
    e1 = np.cross(e2,e3)

    px,py = qxyz.dot(e1),qxyz.dot(e2)
    if v:
        return px,py,e0.dot(e1)
    else:
        return px,py

def project_beams2D(K,qxy):
    u = K/np.linalg.norm(K)
    v = np.array([u[1],-u[0]])
    px = qxy.dot(v)
    return px


def remove_friedel_pairs(reflF):
    hkls = np.array([np.array(h[1:-1].split(','),dtype=int)  for h in reflF])
    refl=[]
    for h,hkl in zip(reflF,hkls):
        if not str(tuple(-hkl)) in refl:
            refl.append(h)
    print('removing Friedel pairs')
    return refl

def show_tiff(file,cutoff=100):
    import tifffile
    im=tifffile.imread(file)
    dsp.stddisp(im=[im],title=file,
        cmap='gray',caxis=[0,cutoff],pOpt='tX')


def to_shelx(hkl,file):
    '''converts to a.hkl file ready to use by shelx
    hkl: dataframe containing h,k,l,I,sig
    '''
    formats = {
        'h':'{:>4}', 'k':'{:>4}', 'l':'{:>4}',
        'I':'{:>8.2f}', 'sig':'{:>8}',
        }
    formatters = {k: v.format for k, v in formats.items()}
    content=hkl.to_string(formatters=formatters, index=False, header=False)
    with open(file, 'w') as f:
        f.write(content)

    print(colors.green+'file saved : \n'+colors.yellow,file,colors.black)

def read_space_group(struct_file):
    dataset={}
    if struct_file.split('.')[-1]=='cif':
        with open(struct_file,'r') as f:
            lines=f.readlines()
        spg_info = {
            'lattice_system'      :['_space_group_crystal_system','_symmetry_cell_setting'],
            'international_number':['_space_group_IT_number'     ,],
            'international_symbol':['_space_group_name_H-M_alt'  ,'_symmetry_space_group_name_H-M'],
        }

        # (key,vals),l=next(spg_info),next(lines)
        for l in lines:
            if l.strip() in ['_space_group_symop_operation_xyz','_symmetry_equiv_pos_as_xyz']:
                break
            for key,vals in spg_info.items():
                for v in vals:
                    if v==l.split(' ')[0]:
                        val = l.replace(v,'').strip(" '\n")
                        # print('match found:',l.split(' ')[0],v,val)
                        dataset[key]=val
                        break

    elif struct_file.split('.')[-1]=='pdb':
        with open(struct_file,'r') as f:
            lines=f.readlines()
        for l in lines:
            if 'SYMMETRY OPERATORS FOR SPACE GROUP' in l:
                dataset['international_symbol'] = l.split(':')[1].strip()
                break

    if dataset:
        print(colors.green+'Reading space group info manually in cif file'+colors.black)
        if 'international_symbol' in dataset.keys() and 'international_number' not in dataset.keys():
            with open('static/spg/spg_groups_hm.json','r') as f:
                spg_hm=json.load(f)
            hm = dataset['international_symbol']
            if hm in spg_hm.keys():
                dataset['international_number'] = spg_hm[hm]
                print(colors.blue+'retrieved space group number from dictionary'+colors.black)

    return dataset



''' Build wallpapers from the 17 plane groups
'''
from utils import displayStandards as dsp
from utils import glob_colors as colors
from matplotlib import patches
from matplotlib import collections as coll
from scipy.spatial import Delaunay
import matplotlib.tri as mtri
import os
import numpy as np
import pandas as pd
from . import lattice
from . import nodesTri as nt
#import lattice
#import nodesTri as nt

pp_types = ['p1','p2','pm','pg','cm','pmm','pmg','pgg','cmm',
            'p4','p4m','p4g',  'p3','p3m1','p31m','p6','p6m']; #print(pp_types)
pp_long  = ['p1','p2','p1m1','p1g1','c1m1','p2mm','p2mg','p2gg','c2mm',
            'p4','p4mm','p4gm', 'p3','p3m1','p31m','p6','p6mm']
pp_lattices = ['oblique']*2+['rect']*7+['square']*3+['hex']*5

df_wallpp = pd.DataFrame(
    np.array([pp_long,pp_lattices,]).T,
    columns = ['long_name','lat_type'],
    index = pp_types)

################################# asym unit cells
u1,u21   = np.array([[0.0,0],[1,0],[1,1],[0,1]]),  np.array([[0,0],[1/2,0],[1/2,1],[0,1]])
u22,u42  = u1/2,u21/2
u41      = np.array([[0,0,],[1/4,0],[1/4,1],[0,1]])
u22x = np.array([[0,0],[1/2,0],[1/2,1/2]])
u22y = np.array([[0,0],[1/2,0],[0,1/2]])
u3   = np.array([[0,0],[1/2,-1/3],[1,0],[1/2,1/3]])
u3m1 = np.array([[0,0],[1/2,-1/3],[1/2,1/3]])
u31m = np.array([[0,0],[1,0],[1/2,1/3]])
u6   = np.array([[0,0],[1/2,-1/3],[1/2,1/3]])
u6m  = np.array([[0,0],[1/2,0],[1/2,1/3]])
asym_cells = {'p1':u1,'p2':u21 ,'pm':u21,'pg':u21 ,'cm':u22 ,'pmm':u22 ,
             'pmg':u41,'pgg':u22,'cmm':u42,'p4':u22,'p4m':u22x,'p4g':u22y,
             'p3':u3,'p3m1':u3m1,'p31m':u31m,'p6':u6,'p6m':u6m}
df_wallpp['asym_cell'] = asym_cells.values()


################################# Atomic parameters
V0 = 1.0
Ai = np.array([0.1,0.25,0.26,0.27,1.5])
fv = lambda X,X0,Za : V0*np.exp(-(np.linalg.norm(X-X0,axis=1)/Ai[Za])**2)
#f  = np.sum(np.array([fc(X,X0) for X0 in Xasym]),axis=0)
g,marker=0.5,'o' #^'
atoms = pd.DataFrame.from_dict(                                         # covalent radius in A
    { 0:['H',(0,0,0),0.37], 1:['C',(g,g,g),0.77],
      2:['N',(0,0,1),0.75], 3:['O',(1,0,0),0.73],4:['S',(0,1,1),1.02]},
    orient='index',columns=['atom','color','size'])


class Wallpaper :
    '''Class for creating wallpapers'''
    def __init__(self,pp_type,a=1,b=1,angle=90,pattern=None,gen=True,ndeg=0,nh=3,nk=3,pOpt='',name='',path='.') :
        ''' initialize class
        pp_type   : plane group type (str)
        a,b,angle : lattice parameters (double)
        pattern   : x y f coordinates within asymetric unit cells   (see add_pattern)
        gen       : generate wallpaper applying symmetry operations (see generate_wallpp for ndeg,nh,nk)
        pOpt      : plot options                                    (see plot_unit_cells)
                  : 'w' to see the unit cell if warning is issued
            Note : gen,ndeg,nh,nk will work only if pattern is defined
        ndeg   : number of discretization points along one the direction for building the potential map
        '''
        self.pp_type  = pp_type
        self.name     = name if name else df_wallpp.loc[pp_type]['long_name']
        self.init_lattice(a,b,angle)
        self.init_unit_cell()
        self.path = path
        self.ndeg = ndeg
        if isinstance(pattern,list):pattern=np.array(pattern)
        if isinstance(pattern,np.ndarray) or isinstance(pattern,str):
            self.add_pattern(pattern,pOpt)
            if gen :
                self.generate_wallpp(nh,nk,ndeg)
                if pOpt and 'q' not in pOpt:
                    self.plot_unit_cells(pOpt,nh,nk)#,xylims=[0,1,0,0.5])

    def init_lattice(self,a,b,angle) :
        '''init lattice assigning attributes :
        lat_type    : lattice type
        lattice_vec : lattice vectors
        a,b,angle   : lattice parameters enforcing lattice constraints
        '''
        self.lat_type = df_wallpp.loc[self.pp_type]['lat_type']
        if self.lat_type == 'hex' :
            self.a,self.b,self.angle = a,a,60
            self.lattice_vec = np.array([[np.cos(np.pi/3),-np.sin(np.pi/3)],[np.cos(np.pi/3),np.sin(np.pi/3)]])*a
        elif self.lat_type == 'square' :
            self.a,self.b,self.angle = a,a,90
            self.lattice_vec = np.array([[1,0],[0,1]])*a
        elif self.lat_type == 'rect' :
            self.a,self.b,self.angle = a,b,90
            self.lattice_vec = np.array([[a,0],[0,b]])
        elif self.lat_type == 'oblique' :
            self.a,self.b,self.angle = a,b,angle
            ang = np.deg2rad(angle)
            self.lattice_vec = np.array([[a,0],[b*np.cos(ang),b*np.sin(ang)]])
        self.params = [self.a,self.b,np.pi/180*self.angle]

    def init_unit_cell(self) :
        '''init unit cell defining :
        - sym_ops   : symmetry operations
        - asym_cell : asymmetric unit cell (matplotlib.patches.Polygon)'''
        self.sym_ops   = wallpp_syms(self.pp_type,self.lattice_vec)
        verts    = asym_cells[self.pp_type]
        a,b,angle = self.params
        if self.lat_type == 'hex' :
            #print('HEX')
            verts[:,0] *=a
            verts[:,1] *=a*np.cos(np.pi/6)
        else :
            verts = self.lattice_vec.T.dot(verts.T).T         # 2x2.2x3 = 2x3
        self.asym_cell = patches.Polygon(verts,alpha=0.3,edgecolor='k',linewidth=1.5,label='$asym_cell$')

    def add_pattern(self,pattern,pOpt='') :
        ''' add a pattern from :
        Nx3 np.array : x y f coordinates within asymetric unit cells
        txt file     : contains  x y f
        where f : list of tags for atoms '''
        if isinstance(pattern,str) : pattern = np.loadtxt(pattern)
        self.X0 = pattern[:,[0,1]]
        self.f0 = pattern[:,2]
        idb = self.asym_cell.contains_points(self.X0)
        if (~idb).sum() :
            print(colors.red+"warning:pattern not in asymetric unit cells"+colors.black);
            pOpt += ['','W']['w' in pOpt]
            print(self.X0);print(self.asym_cell.get_verts());print(idb)
        if 'W' in pOpt:
            a,b = self.lattice_vec
            pcell = patches.Polygon(np.array([[0,0],a,a+b,b]),label='$unit_cell$')
            colls  = [coll.PatchCollection([pcell],alpha=0.1,edgecolor=(g,g,g),linewidth=1),
                      coll.PatchCollection([self.asym_cell],alpha=0.3,edgecolor='k',linewidth=1.5,label='$asym_cell$')]
            (x,y),c = self.X0.T,self.f0
            dsp.stddisp(scat=[x,y,c],colls=colls,opt=pOpt,ms=20,title=self.pp_type,equal=True)
        #self.X0 = X0[idb]

    def generate_wallpp(self,nh,nk,ndeg=0) :
        ''' Generate wallpaper applying symmetry operations :
        ndeg   : order for generating potential map              (see build_potential)
        nh,nk  : number of unit cells to be displayed
        ndeg   : number of discretization points along one the direction for building the potential map'''

        if self.pp_type=='p1' and ndeg:
            self.build_potential_p1(ndeg)
        else:
            self.Xcell,self.fcell   = self.apply_symmetries(self.X0,self.f0)
            self.Xa,self.fa         = self.repeat_pattern(self.Xcell,self.fcell,nh,nk)
            if ndeg :
                self.build_potential(ndeg)
                self.Xc,self.fc = self.apply_symmetries(self.Xp,self.fp)
                self.X,self.f   = self.repeat_pattern(self.Xc,self.fc,nh,nk)

    def build_potential_p1(self,ndeg=2**8):
        print('...building potential...')
        if isinstance(ndeg,int):ndeg=[ndeg]*2
        self.x,self.z = np.linspace(0,self.a,ndeg[0]),np.linspace(0,self.b,ndeg[1])
        x,z = np.meshgrid(self.x,self.z)
        self.Xa,self.fa=self.X0,self.f0
        self.fp=np.zeros(x.shape)
        for X0,Za in zip(self.Xa,self.fa):
            r = np.sqrt((x-X0[0])**2+(z-X0[1])**2)
            self.fp += np.exp(-(r/Ai[int(Za)])**2)
        self.Xp = np.array([x.flatten(),z.flatten()]).T

    def build_potential(self,ndeg=10) :
        ''' Build the potential map :
        ndeg : number of discretization points along one the direction for building the potential map'''
        vert_cell = self.asym_cell.get_verts()[:-1,:]   #last point is sme as first in polygon
        if vert_cell.shape[0]==3 :
            xi,eta = nt.nodesTri(ndeg).T                #get triangular elements
            N  = 0.5*np.array([-xi-eta,1+xi,1+eta]).T   #lin shape functions at xi,eta
            self.Xp  = N.dot(vert_cell)                 #transform : Nx3.3x2=Nx2
        if vert_cell.shape[0]==4 :
            # ndeg+=1
            nlin = np.arange(ndeg)/(ndeg); #np.linspace(0,1,ndeg)
            x,y = np.meshgrid(nlin,nlin)
            N = np.array([x.flatten(),y.flatten()]).T     #get rectangular grid Nx2
            vecs = vert_cell[[1,-1],:]                  #2x2 [a;b]
            self.Xp  = N.dot(vecs)                      #transform : Nx2.2x2=Nx2
        self.Xp += self.lattice_vec[0]+self.lattice_vec[1]
        #print(self.Xp)
        self.fp = np.sum(np.array([fv(self.Xp,X0,int(Za)) for X0,Za in zip(self.Xa,self.fa)]),axis=0)
        self.Xp -= self.lattice_vec[0]+self.lattice_vec[1]
        #self.fp  = fv(self.Xp,self.Xa)                  #potential

    def get_potential_grid(self):
        '''potential on a grid within unit cell
        ADAPT TO P2 etc...'''
        (x,z),f = self.Xc.T,self.fc
        Nxz   = (self.ndeg,self.ndeg)
        x,z,f = x.reshape(Nxz),z.reshape(Nxz),f.reshape(Nxz)
        return x[0,:],z[:,0],f

    def get_potential_grid_p1(self,ndeg=2**8):
        if not 'fp' in self.__dict__.keys():
            self.build_potential_p1(ndeg)
        return self.x,self.z,self.fp

    def plot_unit_cells(self,opts='AV',nh=3,nk=4,lvls=10,alpha=0.5,fg='12',**kwargs) :
        '''Display the wallpaper
        pOpt :  A(Atom), V(potential), u(unit_cell), a(asym_unit)
        '''
        fig,ax = dsp.create_fig(figsize=fg,pad=[2.,5]['t' in opts])
        colls,scat,im,cs,contours=[],[],None,None,None
        lOpt = ''
        if self.lat_type=='hex' : lOpt += 'hq'
        if 'u' in opts : lattice.plot_lattice(self.lattice_vec,max(nh,3),max(nk,3),pOpt=lOpt,ax=ax)
        if 'a' in opts : colls=[coll.PatchCollection([self.asym_cell],alpha=0.3,linewidth=1.5,edgecolor='b')]
        #plot potential
        if 'V' in opts :
            print('display potential')
            # na = int(self.Xcell.shape[0]/self.X0.shape[0])
            N = int(np.sqrt(self.Xp.shape[0]))
            (x,y),f = self.Xp.T, self.fp
            # triang = mtri.Triangulation(x, y, Delaunay(self.Xc).vertices)
            # cs=ax.tricontourf(triang,self.fc,levels=lvls)
            # cont = ax.contourf(x.reshape((N,N)),y.reshape((N,N)),f.reshape((N,N)),levels=lvls)
            im = [x.reshape((N,N)),y.reshape((N,N)),f.reshape((N,N))]#,snap=True)
            # contours = im+[lvls]
            im+=[alpha]
            # cs = ax.pcolormesh(x.reshape((N,N)),y.reshape((N,N)),f.reshape((N,N)))#,snap=True)

        #plot atoms
        if 'A' in opts :
            x,y = self.Xa.T
            s,c = atoms.iloc[self.fa][['size','color']].values.T
            scat=[x.flatten(),y.flatten(),np.int_(50*np.array(s)),c]
        dsp.stddisp(fig=fig,ax=ax,labs=[r'$x(\AA)$',r'$y(\AA)$'],
            scat=scat,colls=colls,im=im,contour=contours,
            title=self.name,name=self.path+self.name,
            **kwargs)

    def get_reciprocal_lattice_2D(self):
        a1,a2 = self.lattice_vec
        b1,b2 = lattice.reciprocal_lattice_2D(a1,a2)
        return b1,b2

    #private (should be)
    def apply_symmetries(self,X0,f0) :
        Xcell = X0
        for sym in self.sym_ops :
            Xcell = np.concatenate((Xcell,sym(X0)))
        fcell = np.tile(f0,len(self.sym_ops)+1)
        return Xcell,fcell

    def _apply_symmetries(self) :
        self.pattern = config.apply_symmetries(self.pp,self.pattern)

    def repeat_pattern(self,Xcell,fcell,nh=3,nk=4) :
        miller,u_hk = lattice.get_miller(self.lattice_vec,nh,nk)
        # u_hk,miller = lattice.get_miller(self.lattice_vec,nh,nk)
        X = np.empty(shape=(0, 2))
        for u in u_hk :
            X = np.concatenate((X,u+Xcell),axis=0)
        f = np.tile(fcell,miller.shape[0])
        #print(X.shape,f.shape)
        return X,np.array(f,dtype=int)



################################################################
#Define symmetry operations
def mirr(x,u,A) :
    ''' Symmetry with respect to line :
    u : direction
    A : Point belonging to the line
    '''
    #print(u,np.linalg.norm(u))
    u = u/np.linalg.norm(u)
    n  = np.array([-u[1],u[0]]) #;print(n);print(x)
    xa = x-np.array(A)
    return x-2*n*xa.dot(n)[:,None]
def glax(x,u,A):
    ''' Glide axis transformation
    u : direction of the glide axis
    A : Point belonging to the line
    '''
    #u = np.linalg.norm(u)
    xM = mirr(x,u,A)
    xg = xM+0.5*u
    return xg

def wallpp_syms(pp_type,lattice_vec):
    cp,sp=np.cos(2*np.pi/3),np.sin(2*np.pi/3)
    a1,a2 = lattice_vec
    #a,b = lat_params[:2]
    O =  np.array([0,0])
    M = (a1+a2)/2; #print(M)
    Pp,Pm = a1+a2,a1-a2
    x0 = 0*np.array([1/2,1/3])
    rot = lambda t: np.array([[np.cos(t),-np.sin(t)],[np.sin(t),np.cos(t)]])
    centr = lambda x:M+x
    invc  = lambda x:Pp-x
    invc2 = lambda x:M-x
    def mir_uA(u,A) : return lambda x:mirr(x,u,A)
    def gla_uA(u,A) : return lambda x:glax(x,u,A)
    mir2 = mir_uA(a2,a1/2)
    def rot3(i) : return lambda x:rot(2*i*np.pi/3).dot(x.T).T
    def rot3m1(i) : return lambda x:rot(2*i*np.pi/3).dot(mirr(x,Pm,O).T).T
    def rot31m(i) : return lambda x:rot(2*i*np.pi/3).dot(mirr(x,Pp,O).T).T
    def rot4(i) : return lambda x:rot(i*np.pi/2).dot((x-M).T).T+M
    def rot6(i) : return lambda x:rot(i*np.pi/3).dot(x.T).T
    def rot6m(i) : return lambda x:rot(i*np.pi/3).dot(mirr(x,a1,O).T).T
    wallpp_sym={
       'p1'  :[],
       'p2'  :[invc],
       'pm'  :[mir_uA(a2,a1/2)],
       'pg'  :[gla_uA(a1,a2/2)],
       'cm'  :[mir_uA(a1,a2/2),centr,gla_uA(a1,a2/4)],
       'pmm' :[invc,mir_uA(a1,a2/2),mir_uA(a2,a1/2)],
       'pmg' :[invc,mir_uA(a2,a1/4),gla_uA(a1,a2/2)],
       'pgg' :[invc,gla_uA(a2,a1/4),gla_uA(a1,a2/4)],
       'cmm' :[invc,mir_uA(a1,a2/2),mir_uA(a2,a1/2),centr,invc2,gla_uA(a2,a1/4),gla_uA(a1,a2/4)],
       'p4'  :[invc,rot4(1),rot4(3)],
       'p4m' :[invc,rot4(1),rot4(3),mir_uA(a1,a2/2),mir_uA(a2,a1/2),mir_uA(Pp,M),mir_uA(Pm,M)],
       'p4g' :[invc,rot4(1),rot4(3),gla_uA(a2,a1/4),gla_uA(a1,a2/4),mir_uA(Pm,M/2),gla_uA(Pp,O)],
       'p3'  :[rot3(1),rot3(2)],
       'p3m1':[rot3(1),rot3(2),rot3m1(0),rot3m1(1) ,rot3m1(2)],
       'p31m':[rot3(1),rot3(2),rot31m(0),rot31m(1) ,rot31m(2)],
       'p6'  :[rot6(i) for i in range(1,6)],
       'p6m' :[rot6(i) for i in range(1,6)]+[rot6m(i) for i in range(6)]
        }[pp_type]
    return wallpp_sym



import importlib as imp
import pandas as pd,numpy as np,easygui
from utils import displayStandards as dsp ;imp.reload(dsp)


def show_frame(opts='Pqr',mag=500,rot=0,hkl_idx=[],single_mode=False,
    df_pets=None,im_multi=None,df_bloch=None,cmap=None,cutoff=100,
    **kwargs):
    ''' Show a frame with information specified by opts
    - opts : P(proc), M(Multislice), B(Bloch), K(kin), V(Vg), S(Sg), L(lat)
        h(hkl_exp),r(rings), k(hkl_kin) l(legend)
    '''
    labs,qmax,wm = ['$q_x$','$q_y$'],0,0.05
    plts,scat,txts,im,legElt = [[0,0,'b+','']],(),[],None,None

    caxis=[0,cutoff]
    if not cmap:
        cmap='Reds'
        if single_mode  :
            v = [c in opts for c in 'PMBKVSL']
            k = np.array(list(cmaps.keys()))[v][0]
            cmap = cmaps[k]
    # print(single_mode,cmap)

    if 'P' in opts and isinstance(df_pets,pd.DataFrame):
        qx_p,qy_p,I = df_pets[['qx','qy','I']].values.T
        if qx_p.size:
            c = cp
            if single_mode:
                c = I
            scat += ([qx_p,qy_p,I,c,mp],)
            qmax = np.ceil(max(abs(qx_p).max(),abs(qy_p).max()))

            if 'h' in opts:
                hx,kx,lx = df_pets[['hx','kx','lx']].values.T
                sx = lambda x:['%d' %round(x),'%.1f' %x][abs(x-round(x))>0.06]
                txts += [[x+0*wm,y+wm,'(%s,%s,%s)' %(sx(h0),sx(k0),sx(l0)),(0.5,)*3] for x,y,h0,k0,l0 in zip(qx_p,qy_p,hx,kx,lx)]

    if isinstance(df_bloch,pd.DataFrame):
        qx_b,qy_b = df_bloch[['px','py']].values.T
        if rot:
            ct,st = np.cos(np.deg2rad(rot)),np.sin(np.deg2rad(rot))
            qx_b,qy_b = ct*qx_b-st*qy_b,st*qx_b+ct*qy_b
        for c,C in dfb.iterrows() :
            if c in opts:
                I = C.fz(df_bloch[C.F])
                vals = I#/I.max()
                # cs = C.color
                cs=vals
                cmap_k=cmaps[c]
                if single_mode:cmap_k = cmap
                sargs = {'vmin':0.9*vals.min(),'vmax':vals.max(),'cmap':cmap_k}
                scat += ( [qx_b,qy_b,vals/vals.max()*mag,cs,C.marker,sargs],)
        bs = [c for c in dfb.index if c in opts]
        if 'k' in opts and bs and any(hkl_idx):
            h,k,l,qx0,qy0 = df_bloch.loc[hkl_idx][['h','k','l','px','py']].values.T
            if rot:qx0,qy0 = ct*qx0-st*qy0,st*qx0+ct*qy0
            ctxt = dfb.loc[bs[0]].color
            txts += [[x,y+wm,'(%d,%d,%d)' %(h0,k0,l0),ctxt] for x,y,h0,k0,l0 in zip(qx0,qy0,h,k,l)] #,I_b) if I>Itol]
        qmax = np.ceil(max(qmax,qx_b.max(),qy_b.max()))

    if 'M' in opts and im_multi:
        qx_m,qy_m,I=im_multi
        qmax = np.ceil(max(qmax,qx_m.max(),qy_m.max()))
        im=[qx_m,-qy_m,I]

    if 'r' in opts:
        t = np.linspace(0,np.pi*2,100)
        ct,st = np.cos(t),np.sin(t)
        plts+=[[i*ct,i*st,'m','',0.5] for i in np.arange(0.25,qmax,0.25)]
        plts+=[[i*ct,i*st,'m','',2] for i in np.arange(1,qmax)]

    if 'l' in opts:
        legElt = {k:v for c,(k,v) in zip('PMBKVSL',legElts.items()) if c in opts}

    if not 'fonts' in kwargs.keys():kwargs['fonts'] = {'text':15}
    if not 'xylims' in kwargs.keys():kwargs['xylims'] = qmax
    return dsp.stddisp(plts,scat=scat,im=im,texts=txts,bgcol=None,gridOn=0,
        labs=labs,cmap=cmap,caxis=caxis,legElt=legElt,**kwargs)

########################################################################
#### misc
########################################################################
def get_fz(opts):
    '''mapping for function to apply
    - opts : 'r'(real) 'i'(imag) 'a'(angle) 'm'(mag) 'l'(log10(|Fhkl|+1)) '2'(mag^2) 'L'(logM)
    '''
    if not opts:opts='m'
    keys   =['r','i','a','m','l','2','L']
    fs     = [np.real,np.imag,np.angle,np.abs,logF,abs2,logM]
    fs_str = ['real part','imag part','phase','magnitude','$\log_{10}(|F|+1)$','$|F|^2$','$-\log_{10}$']
    fz     = dict(zip(keys,fs))[opts]
    fz_str = dict(zip(keys,fs_str))[opts]
    return fz,fz_str

abs2 = lambda F:np.abs(F)**2
logF = lambda F:np.log10(np.maximum(np.abs(F),1e-6))+6
logM = lambda F:-np.log10(np.maximum(abs(F),1e-5))

cmaps = {'P':'Greys','M':'Reds','B':'Blues','K':'Greens','V':'Purples','S':'Oranges','L':'Greys'}
dfb = { 'B':['I' ,get_fz('m')[0],dsp.getCs(cmaps['B'],3)[1],'o','Intensity Bloch $I_{g}$'],
        'K':['Ig',get_fz('l')[0],dsp.getCs(cmaps['K'],3)[1],'o','Intensity kinematic $I_{g,kin}$'],
        'V':['Vg',get_fz('m')[0],dsp.getCs(cmaps['V'],3)[1],'d','Potential $V_{g}$'],
        'S':['Sw',get_fz('L')[0],dsp.getCs(cmaps['S'],3)[1],'<','Excitation error $\zeta_{g}$'],
        'L':['L' ,get_fz('m')[0],dsp.getCs(cmaps['L'],3)[1],'s','Lattice']}
dfb = pd.DataFrame.from_dict(dfb,orient='index',columns=['F','fz','color','marker','legend'])
cp,mp = (0.5,)*3,'o'
legElts={'Processed':[cp,mp],'Multislice':'r*'}
legElts.update({C.legend:[C.color,C.marker] for c,C in dfb.iterrows()})

def multenterbox(msg,title,fieldValues,fieldNames):
    fieldValues = easygui.multenterbox(msg, title, fieldNames,fieldValues)
    while True:
        if fieldValues is None:
            break
        errs = list()
        for n, v in zip(fieldNames, fieldValues):
            if v.strip() == "":errs.append('"{}" is a required field.'.format(n))
        if not len(errs):
            break
        fieldValues = easygui.multenterbox("\n".join(errs), title, fieldNames, fieldValues)
    if fieldValues:
        return dict(zip(fieldNames,fieldValues))



import numpy as np
from subprocess import Popen
import utils.physicsConstants as cst
import utils.displayStandards as dsp
import utils.glob_colors as colors

nearBragg_bin=dsp.get_figpath(__file__,'/../nearBragg/bin/')+'nearBragg' #;print(nearBragg_bin)
dtype = np.float128

Cs = [colors.unicolor(0.75),(1,0,0),colors.unicolor(0.25),(0,0,1),(1,0,0),(1,1,0)]
Zs = [1,3,6,7,8,16]
# Ai = np.array([1,2,2,2,3])*0.25
Ai = np.array([0.1,0.25,0.26,0.27,1.5])
# Ai =(10-0.5*np.arange(len(Zs)))/8
# self.fj = lambda ti,j:np.sqrt(np.pi)/A[i]*np.exp(-(np.pi*Ai[j]*q)**2)

class NearBragg():
    ''' Near Bragg (distances in A)
    - pattern : x,z,f (in fractional coordinates)
    - keV  : energy(keV) prevails over lam
    - lam  : wavelength(A)
    - Nx,Nz : number of unit cells in each direction
    ## detector : \n
    - z0   : float - distance of detector to crystal origin
    - npx  : float - number of pixels
    - tmax : float - max half angle (deg)
    - qmax : float - max resolution A^-1  (prevails on tmax if defined)
    OR
    - q0s  : np.ndarray - reciprocal space sampling(A^-1)
    ## misc :\n
    - method :
        - Greens2    : Double scattering
        - Greens     : single scattering exact distance calculations
        - Fresnel    : single scattering Fresnel regime
        - Fraunhofer : Single scattering Fraunhofer regime
        - Holton : call Holton code
    - fjopt : with(1) or without(0) form factor (default 1)
    '''
    def __init__(self,pattern,ax,bz,keV=200,lam=None,path='',
            Nx=1,Nz=1,eps=1,
            z0=1e10,npx=4096,tmax=5,qmax=None,q0s=None,
            method='D',fjopt=1,iZv=5):
        if keV : lam = cst.keV2lam(keV)
        self.pattern = pattern
        self.lam  = dtype(lam)

        #initialization
        self.replicate(ax,bz,Nx,Nz)
        if isinstance(q0s,np.ndarray) :
            self.set_pixels(q0s,z0)
        else :
            self.set_detector(npx,z0,tmax,qmax)
        self.I = np.zeros(self.x0s.shape,dtype=dtype)

        #form factor
        if fjopt==1 :
            self.fj = lambda ti,j:eps*np.sqrt(np.pi)/Ai[j]*np.exp(-(np.pi*Ai[j]*(np.sin(ti)/self.lam))**2)
        elif fjopt==0 :
            #no form factor for comparison with Holton code
            self.fj = lambda ti,j:np.ones(ti.shape)
        #compute
        if   method=='Greens2' or  method=='D' : self._Greens2(iZv)
        elif method=='Proba'   or  method=='P' : self._Proba(iZv)
        elif method=='Greens'  or  method=='G' : self._Greens()
        elif method=='Fresnel'                 : self._Fresnel()
        elif method=='Fraunhofer'              : self._Fraunhofer()
        elif method=='Holton'                  : self._Holton(path=path)

    ################################################################
    # init
    ################################################################
    def replicate(self,ax,bz,Nx,Nz):
        '''replicate unit cell pattern Nx,Nz times'''
        self.Nx,self.Nz = Nx,Nz
        self.ax,self.bz = ax,bz
        x,z,Za = self.pattern
        #arange
        nx = np.arange(-int(Nx/2),int(np.ceil(Nx/2)))
        nx,nz = np.meshgrid(nx,np.arange(Nz))
        nx,nz = nx.flatten(),nz.flatten()
        #replicate
        x  = np.hstack([ (x+i)*self.ax for i in nx])
        z  = np.hstack([ (z+i)*self.bz for i in nz])
        Za = np.hstack([Za]*self.Nx*self.Nz)
        #sort
        idx = np.argsort(np.array(z))
        x,z,Za = x[idx],z[idx],Za[idx]
        #store
        self.x  = np.array(x,dtype=dtype)
        self.z  = np.array(z,dtype=dtype)
        self.Za = np.array(Za,dtype=int)

    def set_detector(self,npx,z0,tmax,qmax=None):
        ''' compute pixel positions and sizes from max scattering angle and distance to sample
        '''
        if qmax:tmax=qmax*self.lam
        self.npx  = npx
        self.z0   = dtype(z0)
        self.tmax  = tmax
        self.x_max = self.tmax*self.z0
        self.px    = self.tmax*self.z0*cst.A/cst.mm/(self.npx-1)

        x0s = np.linspace(-self.x_max,self.x_max,self.npx)
        self.x0s = np.array(x0s,dtype=dtype)+(x0s[1]-x0s[0])/2
        self.q0s = np.array(x0s/self.z0s/self.lam,dtype=dtype)

    def set_pixels(self,q0s,z0):
        self.npx  = q0s.size
        self.tmax = q0s.max()*self.lam
        self.z0   = dtype(z0)
        self.q0s = np.array(q0s,dtype=dtype)
        self.x0s = np.array(self.z0*self.lam*q0s,dtype=dtype)
        self.x_max = self.x0s.max()

    ################################################################
    # display
    ################################################################
    def Pattern_show(self,**kwargs):
        '''Show diffraction pattern'''
        C = [Cs[Z] for Z in self.Za]
        Nx0,Nx1 = int(self.Nx/2),int(np.ceil(self.Nx/2))
        dsp.stddisp(scat=[self.z,self.x,C],
            ms=20,labs=[r'$z(\AA)$',r'$x(\AA)$'],
            xyTicks=[self.bz,self.ax],xylims=[0,self.Nz*self.bz,-Nx0*self.ax,Nx1*self.ax],
            **kwargs)

    # def stat_show(self,**kwargs):
    #     '''show statistics'''
    #     z = self.bz*np.arange(self.Nz+1)/10
    #     cs = dsp.getCs('Spectral',2)
    #     S = np.array([np.cumsum(Si) for Si in self.S])
    #     plts  = [[z,Si,cs[i],'%d' %i] for i,Si in enumerate(S)]
    #     plts += [[z,S.sum(axis=0),'k--']]
    #     return dsp.stddisp(plts,labs=['$z(nm)$','$Proba$'],
    #         **kwargs)

    def F_show(self,qopt=0,**kwargs):
        '''Show atomic form factors'''
        Za = np.unique(self.Za)
        t = self.getAngle()*np.pi/180
        if qopt:
            labx,x = r'$q(\AA^{-1})$',self.getQ()
        else:
            labx,x = r'$\theta (deg)$', self.getAngle()

        plts = [[x,self.fj(t,Z)**2,'b','$Z_a=%d$' %Z] for Z in Za]
        plts += [[x,self.fj(t,Z),'g--','$Z_a=%d$' %Z] for Z in Za]
        dsp.stddisp(plts,labs=[labx,r'$f_j^2(\AA)$'],**kwargs)


    def getAngle(self):return (self.x0s/self.z0)*180/np.pi
    def getQ(self):return (self.x0s/self.z0)/self.lam
    def getI(self):return self.I

    ################################################################
    # utils to run James Holton .c code
    ################################################################
    def _save_input(self,file):
        np.savetxt(file,np.vstack([self.z,np.zeros(self.x.shape),self.x,
            np.ones(self.x.shape),np.zeros(self.x.shape),np.zeros(self.x.shape)]).T)
        print(colors.yellow+file+colors.black)

    def _cmd(self,opts='s',path='',file='atoms.txt'):
        self.path = path
        self.file = file
        if 's' in opts :self._save_input(path+self.file)
        #cmd
        cmd=" %s -file %s " %(nearBragg_bin,path+self.file)
        cmd+='-lambda %f ' %(self.lam)
        cmd+='-distance %f -source_dist far ' %(self.z0*cst.A/cst.mm)
        cmd+= '-pixel %.15f -detpixels_x %d ' %(2*self.px,self.npx)
        cmd+='-detpixels_y 1 -oversample 1 '
        cmd+='-intfile %s -floatfile %s -Ifilename %s ' %(path+'intimage.img',path+'floatimage.bin',path+'I.txt')
        # cmd+='-debug -printout'
        print(colors.red+cmd+colors.black)
        if 'r' in opts:
            print(colors.green+'.........Near Bragg.......'+colors.black)
            p = Popen(cmd,shell=True);p.wait()
            self.I = np.loadtxt(path+'I.txt')
            # print(colors.green+'Near Bragg'+colors.black)

    def _Holton(self,path=''):
        '''Run James Holton code'''
        self._cmd(opts='sr',path=path,file='atoms.txt')

    ################################################################
    # Single scattering routines
    ################################################################
    def _Fraunhofer(self):
        print(colors.green+'... Running nearBragg Fraunhofer ...'+colors.black)
        for i in range(self.npx) :
            tij  = np.abs(self.x0s[i]-self.x)/(self.z0-self.z)
            R_ij = self.z + np.sqrt((self.x0s[i]-self.x)**2+(self.z0-self.z)**2)
            Rij  = self.x0s[i]*self.x/self.z0
            self.I[i] = np.abs(np.sum(
                self.fj(tij,self.Za)*np.exp(2*np.pi*1J*Rij/self.lam)/(R_ij*cst.A)))**2

    def _Fresnel(self):
        print(colors.green+'... Running nearBragg Fresnel ...'+colors.black)
        for i in range(self.npx) :
            tij  = np.abs(self.x0s[i]-self.x)/(self.z0-self.z)
            R_ij = self.z + np.sqrt((self.x0s[i]-self.x)**2+(self.z0-self.z)**2)
            Rij  = (self.x0s[i]-self.x)**2/(2*(self.z0-self.z))
            self.I[i] = np.abs(np.sum(
                self.fj(tij,self.Za)*np.exp(2*np.pi*1J*Rij/self.lam)/(R_ij*cst.A)))**2

    # def _Dual_beams(self):
    #     print(colors.green+'... Running nearBragg Dual Beams ...'+colors.black)
    #     natoms = self.z.size
    #     self.A = np.zeros((self.x0s.size),dtype=np.complex256)
    #     for i in range(natoms) :
    #         if not i%100 : print('%d/%d' %(i,natoms))
    #         #single scattering
    #         tij0  = np.abs(self.x0s-self.x[i])/(self.z0-self.z[i])
    #         Rij0 = self.z[i] + np.sqrt((self.x0s-self.x[i])**2+(self.z0-self.z[i])**2)
    #         self.A += self.fj(tij0,self.Za[i])*np.exp(2*np.pi*1J*Rij0/self.lam)/(Rij0*cst.A)
    #         #double scattering
    #         for j in range(i):
    #             tij = np.arctan((self.x[i]-self.x[j])/(self.z[i]-self.z[j]))
    #             Rij = np.sqrt((self.x[j]-self.x[i])**2+(self.z[j]-self.z[i])**2)
    #             fij = self.fj(tij,self.Za[j]) #/np.sqrt(Rij)
    #             self.A += fij*self.fj(tij0-tij,self.Za[i])*np.exp(2*np.pi*1J*(Rij0+Rij)/self.lam)/(Rij0*cst.A)
    #     self.I = np.abs(self.A)**2

    def _Greens(self):
        print(colors.green+'... Running nearBragg Greens ...'+colors.black)
        for i in range(self.npx) :
            tij  = np.abs(self.x0s[i]-self.x)/(self.z0-self.z)
            R_ij = self.z + np.sqrt((self.x0s[i]-self.x)**2+(self.z0-self.z)**2)
            self.I[i] = np.abs(np.sum(
                self.fj(tij,self.Za)*np.exp(2*np.pi*1J*R_ij/self.lam)/(R_ij*cst.A)))**2

    #################################################################
    #### Multiple scattering routines
    #################################################################
    # def _Greens2n(self,iZv=5):
    #     '''2-level dynamical scattering ignoring backward scattering'''
    #     print(colors.green+'... Running nearBragg Greens2 ...'+colors.black)
    #
    #     ## Primary scattering
    #     ##atoms x pixels
    #     xjp,zjp = self.x0s-self.x[:,None]),self.z0-self.z[:,None]
    #     tjp = np.arctan2(xjp,zjp)
    #     Rjp = np.sqrt(xjp*2+zjp**2)
    #     ej  = np.exp(2J*np.pi*k0*self.z)
    #     Fj  = np.sum(ej*self.fj(tij,self.Za)*np.exp(2J*np.pi*k0*Rjp),axis=1)
    #
    #     ## Secondary scattering
    #     # natoms x natoms
    #     xij,zij,Zi = self.x[:,None]-self.x,self.z[:,None]-self.z,self.Za[:,None]
    #     tij = np.arctan2(xij,zij)
    #     tijp = tjp[:,None] - tij
    #     Rij = np.sqrt(xij*2+zij**2)
    #     Fij = self.fj(tij,self.Za)*self.fj(tijp,Zi)*np.exp(2J*np.pi*k0*Rjp[:,None])

    def _Greens2(self,iZv=5):
        '''2-level dynamical scattering ignoring backward scattering'''
        print(colors.green+'... Running nearBragg Greens2 ...'+colors.black)
        natoms  = self.z.size
        dq      = self.q0s[1]-self.q0s[0]
        # sig_e   = sum(self.fj(self.q0s*self.lam,self.Za[0])**2)*dq
        # print('dq=%.1E Angstrom, sig_e:%.1E Angstrom' %(dq,sig_e))

        I0 = 1
        self.A = np.zeros((self.x0s.size),dtype=complex)
        Adyn = np.zeros((self.x0s.size),dtype=complex)
        # self.S[0,0]=1
        iz=1
        print(colors.yellow+'atom %d/%d slice %d, I0=%.4f' %(1,natoms, iz,I0) +colors.black)
        for i in range(natoms) :
            #### single scattering
            ti0 = (self.x0s-self.x[i])/(self.z0-self.z[i])
            #distance from atom to detector
            Ri0 = np.sqrt((self.x0s-self.x[i])**2+(self.z0-self.z[i])**2)
            # single scattering amplitude
            Akin = self.fj(ti0,self.Za[i])*np.exp(2*np.pi*1J*self.z[i]/self.lam)
            #### double scattering with backward atoms
            ibackward = self.Nx*(iz-1)
            Adyn[:]=0
            for j in range(ibackward):
                #angle between atoms
                tij = np.arctan((self.x[i]-self.x[j])/(self.z[i]-self.z[j]))
                #distance betwwen atoms
                Rij = np.sqrt((self.x[j]-self.x[i])**2+(self.z[j]-self.z[i])**2)
                #amplitude of radiation j at atom i (dimensionless)
                fij = self.fj(tij,self.Za[j])*np.exp(2*np.pi*1J*Rij/self.lam)#*dq
                #scattering amplitude from atom i due to scattering from atom j
                Adyn += fij*self.fj(ti0-tij,self.Za[i])

            # increment slice
            # print(abs(Adyn/Akin).max())
            if not (i+1)%self.Nx:
                # I0 -= self.S[1,iz]/(self.ax*self.Nx) #;print(I0)
                iz+=1
                if not iz%iZv :
                    print(colors.yellow+'atom %d/%d slice %d, I0=%.4f' %(i+1,natoms, iz,I0) +colors.black)

            # atom contribution to diffraction pattern at detector
            # print(np.abs(Akin).max(),np.abs(Adyn).max(),np.abs(Akin-Adyn).min())
            self.A += (Akin+Adyn)*np.exp(2*np.pi*1J*Ri0/self.lam)/(Ri0*cst.A) #*np.sqrt(I0)
        self.I = np.abs(self.A)**2

    # def _Greens2_old(self,iZv=5):
    #     print(colors.green+'... Running nearBragg Greens2 ...'+colors.black)
    #     natoms  = self.z.size
    #     dq      = self.q0s[1]-self.q0s[0]
    #     sig_e   = sum(self.fj(self.q0s*self.lam,self.Za[0])**2)*dq
    #     print('dq=%.1E Angstrom, sig_e:%.1E Angstrom' %(dq,sig_e))
    #
    #     I0 = 1/(self.ax*self.Nx) #1 electron/transverse area/sec
    #     self.A = np.zeros((self.x0s.size),dtype=complex)
    #     self.S = np.zeros((2,self.Nz+1))
    #     self.S[0,0]=1
    #     iz=1
    #     for i in range(natoms) :
    #         if not i%(iZv*self.Nx) : print('atom %d/%d slice %d, I0=%.4f' %(i+1,natoms, iz,I0))
    #         #### single scattering
    #         tij0 = (self.x0s-self.x[i])/(self.z0-self.z[i])
    #         R_ij = self.z[i] + np.sqrt((self.x0s-self.x[i])**2+(self.z0-self.z[i])**2)
    #         Akin = np.sqrt(I0)*self.fj(tij0,self.Za[i])*np.exp(2*np.pi*1J*R_ij/self.lam) #/(R_ij*cst.A)
    #
    #         Skin = (abs(Akin)**2).sum()*dq
    #         assert Skin<1
    #         self.S[0,iz] -= sig_e*I0
    #         self.S[1,iz] += Skin
    #         self.A += Akin
    #         if not (i+1)%self.Nx:
    #             I0 -= self.S[1,iz]/(self.ax*self.Nx) #;print(I0)
    #             iz+=1
    #     self.I = np.abs(self.A)**2

    ##
    def _Proba(self,iZv=5):
        print(colors.green+'... Running nearBragg Proba ...'+colors.black)
        natoms = self.z.size
        dq      = self.q0s[1]-self.q0s[0]
        sig_e   = sum(self.fj(self.q0s*self.lam,self.Za[0])**2)*dq
        self.sig_e = sig_e
        print('dq=%.1E Angstrom, sig_e:%.3f Angstrom' %(dq,sig_e))

        I0 = 1/(self.ax*self.Nx) #1 electron/transverse area/sec
        # self.S = np.zeros((self.Nz+1,self.x0s.size,3)) #scattering areas
        self.I = np.zeros((3,self.Nz+1))
        self.S,Iz = np.zeros((3,self.Nz+1)),np.zeros((self.Nz+1))
        self.I[0,0],self.S[0,0],Iz[0]=1,1,I0
        iz=1
        for i in range(natoms) :
            if not i%(iZv*self.Nx) : print('atom %d/%d slice %d' %(i+1,natoms, iz))
            #### single scattering
            tij0  = (self.x0s-self.x[i])/(self.z0-self.z[i])

            self.S[0,iz] -= sig_e*I0 #fraction electron scattered per unit time
            self.S[1,iz] += np.sum(self.fj(tij0,self.Za[i])**2)*dq*I0
            #### double scattering
            # print(colors.green,i,colors.black)
            ibackward = self.Nx*(iz-1)
            for j in range(ibackward):
                tij = np.arctan((self.x[i]-self.x[j])/(self.z[i]-self.z[j]))
                if abs(tij*180/np.pi)<4 : #and abs(tij)>1e-5:
                    fij = 1#self.fj(tij,self.Za[j])*dq
                    fj  = np.sum((fij*self.fj(tij0-tij,self.Za[i]))**2)*dq
                    jz  = int(j/self.Nx)+1
                    self.S[1,jz] -= fj*self.S[1,jz]/(self.ax*self.Nx)#;Iz[jz]
                    self.S[2,iz] += fj*self.S[1,jz]/(self.ax*self.Nx)

                    # iq  = np.argmin(np.abs(self.q0s-tij/self.lam))
                    # jz  = int(j/self.Nx)+1
                    # self.S[iz,:, 2] += fj*I0
                    # self.S[jz,iq,1] -= fj.sum()*dq*I0
                    # print(colors.red+'%3d' %j+colors.black+' : %.3f,%2d,%.2E,%.2E' %(tij*180/np.pi,iq,fij,fj.sum()*dq))
            if not (i+1)%self.Nx:
                I0 += self.S[0,iz]/(self.ax*self.Nx) #;print(I0)
                Iz[iz] = I0
                self.I[0,iz] = self.I[0,iz-1]+self.S[0,iz]
                self.I[1,iz] = np.sum(self.S[1,:])
                self.I[2,iz] = self.I[2,iz-1]+self.S[2,iz]
                iz+=1


        # self.I[0,:] = np.cumsum(self.S[0,:])
        # self.I[1,:] = np.cumsum(self.S[1,:])
        # self.I[2,:] = np.cumsum(self.S[2,:])
        # self.I[0,:] = 1 - np.cumsum(self.S[:,0,0])
        # self.I[1,:] = np.cumsum(self.S[:,:,1].sum(axis=1))*dq
        # self.I[2,:] = np.cumsum(self.S[:,:,2].sum(axis=1))*dq

####################################################################
# misc
####################################################################
disp_d ={
    'q':('getQ'     ,'k',r'$q(\AA^{-1})$'   ),
    'a':('getAngle' ,'k',r'$\theta(\circ)$' ),
    'I':('getI'     ,'r',r'$I$' ),
}

def display(objs,key_pair,**kwargs):
    return dsp.display_solutions(disp_d,objs,key_pair)
